(*+--------------------------------------------------------------------------+
  |                                                                          |
  |                         valuation d'une expression                      |
  |                                                                          |
  +--------------------------------------------------------------------------+*)

type valeur = Int of int | Fct of mmoire * string * expression | Undef
and mmoire == (string * (valeur ref)) list;;
(* valeur ref car il faut pouvoir modifier l'association bidon en cas *)
(* de dfinition rcursive                                            *)

(* oprations binaires *)
let opration = fun
  | (Int(a), "+" , Int(b)) -> Int(a+b)
  | (Int(a), "-" , Int(b)) -> Int(a-b)
  | (Int(a), "*" , Int(b)) -> Int(a*b)
  | (Int(a), "/" , Int(b)) -> Int(a/b)
  | (Int(a), "=" , Int(b)) -> Int(if a =  b then 1 else 0)
  | (Int(a), "<" , Int(b)) -> Int(if a <  b then 1 else 0)
  | (Int(a), ">" , Int(b)) -> Int(if a >  b then 1 else 0)
  | (Int(a), "<=", Int(b)) -> Int(if a <= b then 1 else 0)
  | (Int(a), ">=", Int(b)) -> Int(if a >= b then 1 else 0)
  | (Int(a), "<>", Int(b)) -> Int(if a <> b then 1 else 0)
  | _ -> failwith "opration non valuable"
;;

(* valuation *)
let rec value mmoire exp = match exp with
| Const(c)   -> Int(c)
| Nom(x)     -> !(assoc x mmoire)
| Fun(x,e)   -> Fct(mmoire,x,e)
| Bin(a,o,b) -> opration(value mmoire a, o, value mmoire b)
| Appl(f,x)  ->
    begin match value mmoire f with
    | Fct(m,t,e) -> value ((t,ref(value mmoire x)) :: m) e
    | _          -> failwith "expression non fonctionnelle"
    end
| If(c,v,f) ->
    begin match value mmoire c with
    | Int(0) -> value mmoire f
    | Int(_) -> value mmoire v
    | _      -> failwith "expression non boolenne"
    end
| Letin(x,e,f) -> (* dclare x en mmoire en cas de dfinition rcursive *)
    let rx = ref(Undef) in
    let mem = (x,rx) :: mmoire in
    rx := value mem e;
    value mem f
;;

(* excution *)
let mem_globale = ref [];;

let excute texte = match arbre(mots(texte)) with
| E(e)     -> value !mem_globale e
| Def(x,e) -> (* dclare x en mmoire en cas de dfinition rcursive *)
    let rx = ref(Undef) in
    let mem = (x,rx) :: !mem_globale in
    rx := value mem e;
    mem_globale := mem;
    !rx
| _ -> failwith "cas impossible"
;;

(* imprimeur de valeur (pour viter les impressions rcursives) *)
#open "format";;
let print_valeur = function
| Int(n)     -> print_string "Int "; print_int n
| Fct(m,x,e) -> print_exp (Fun(x,e))
| Undef      -> print_string "Undef"
;;
#close "format";;
install_printer "print_valeur";;

