(*+--------------------------------------------------------------------------+
  |                                                                          |
  |                           Excution du code compil                      |
  |                                                                          |
  +--------------------------------------------------------------------------+*)

#open "printf";;

     (*+------------------------------------------------------+
       |  Mmoire : les premiers mots simulent les registres  |
       +------------------------------------------------------+*)

let mmoire = make_vect 1000 (Int(0));;
let reset() =
  mmoire.(0) <- Ptr(0);       (* pc *)
  mmoire.(1) <- Ptr(4);       (* sp *)
  mmoire.(2) <- Ptr(1000);    (* hp *)
  mmoire.(3) <- Ptr(1000);    (* ep *)
  mmoire.(4) <- Int(0)	       (* dr *)
;;

                    (*+-------------------------+
                      |  Fonctions d'affichage  |
                      +-------------------------+*)

let prmem ch a = match mmoire.(a) with
| Int(i) -> fprintf ch "Int %d\n" i
| Ptr(p) -> fprintf ch "Ptr %d\n" p
;;

let print_mmoire ch a b =
  for i=a to b do fprintf ch "%3d: " i; prmem ch i done
;;

let print_regs ch =
  fprintf ch "PC: "; prmem ch 0;
  fprintf ch "SP: "; prmem ch 1;
  fprintf ch "HP: "; prmem ch 2;
  fprintf ch "EP: "; prmem ch 3;
  fprintf ch "DR: "; prmem ch 4
;;

                  (*+----------------------------+
                    |  Simulation de la machine  |
                    +----------------------------+*)

let unptr(a) = match mmoire.(a) with
| Ptr(p) -> p
| Int(_) -> failwith "confusion entier pointeur"
;;

let numreg(r) = match r with
| PC -> 0
| SP -> 1
| HP -> 2
| EP -> 3
| DR -> 4
;;

let adr(a) = match a with
| Reg(r)   -> numreg(r)
| Mem(r,i) -> unptr(numreg r)+i
;;

let addreg(r,i) =
  let r=numreg(r) in
  match mmoire.(r) with
  | Ptr(p) -> mmoire.(r) <- Ptr(p+i)
  | Int(p) -> mmoire.(r) <- Int(p+i)
;;

let opration = function
  | (Int(a), "+",  Int(b)) -> Int(a+b)
  | (Int(a), "-",  Int(b)) -> Int(a-b)
  | (Int(a), "*",  Int(b)) -> Int(a*b)
  | (Int(a), "/",  Int(b)) -> Int(a/b)
  | (Int(a), "=" , Int(b)) -> Int(if a =  b then 1 else 0)
  | (Int(a), "<" , Int(b)) -> Int(if a <  b then 1 else 0)
  | (Int(a), ">" , Int(b)) -> Int(if a >  b then 1 else 0)
  | (Int(a), "<=", Int(b)) -> Int(if a <= b then 1 else 0)
  | (Int(a), ">=", Int(b)) -> Int(if a >= b then 1 else 0)
  | (Int(a), "<>", Int(b)) -> Int(if a <> b then 1 else 0)
  | _ -> failwith "opration non valuable"
;;

exception HALT;;

let rec exec(inst) = match inst with
| Load(a,d)   -> mmoire.(adr a) <- d
| Move(a,b)   -> mmoire.(adr a) <- mmoire.(adr b)
| Push(a)     -> do_list exec [Move(Mem(SP,1),a); Add(SP,1)] 
| Pop(a)      -> do_list exec [Move(a,Mem(SP,0)); Add(SP,-1)]
| Op(o,a,b,c) -> mmoire.(adr a) <- opration(mmoire.(adr b),o,mmoire.(adr c))
| Add(r,i)    -> addreg(r,i)
| Call(a)     -> do_list exec [Push(Reg(PC)); Move(Reg(PC),a); Add(PC,-1)]
| Ret         -> exec (Pop(Reg(PC)))
| Jfalse(a,i) -> begin match mmoire.(adr a) with
		   | Int(0) -> mmoire.(0) <- Ptr(i-1)
		   | Int(_) -> ()
		   | Ptr(_) -> failwith "confusion entier pointeur"
		 end
| Jptr(a,i)   -> begin match mmoire.(adr a) with
		   | Ptr(_) -> mmoire.(0) <- Ptr(i-1)
		   | Int(_) -> ()
		 end
| Jmp(i)      -> mmoire.(0) <- Ptr(i-1)
| Print(a)    -> prmem stdout (adr a); flush stdout
| Comm(_)     -> ()
| Halt        -> raise HALT
;;

let excute(adr) =
  mmoire.(0) <- Ptr(adr);
  try while true do exec(code.(unptr 0)); addreg(PC,1) done
  with HALT -> ()
;;

                     (*+-----------------------+
                       |  Interface graphique  |
                       +-----------------------+*)

#open "unix";;
let show() =

  (* lance wish et met en place les listes et boutons *)
  let (chin,chout) = open_process "wish -name machine" in
  fprintf chout "source machine.tk\n";
  fprintf chout "download code\n";
  dsassemble chout 0 !pc;
  fprintf chout "\n";

  (* excute une instruction et affiche l'erreur dans une boite de dialogue *)
  let exec(a) = try exec(a)
  with (Failure msg) as exc ->
    fprintf chout "erreur {%s}\n" msg;
    flush chout;
    let _ = close_process (chin,chout) in
    raise exc
  in

  (* dmarre l'excution *)
  reset();
  let encore = ref(true) in
  while !encore do

    (* affiche les registres, la pile et le tas *)
    fprintf chout "download pile\n";
    print_mmoire chout 5 (unptr 1);
    fprintf chout "\n";
    fprintf chout "download tas\n";
    print_mmoire chout (unptr 2) 999;
    fprintf chout "\n";
    fprintf chout "download reg\n";
    print_regs chout;
    fprintf chout "\n";
    fprintf chout "hilight code %d" (unptr 0);
    fprintf chout "\n";
    fprintf chout ".pile.list see end\n";
    flush chout;

    (* attend un vnement *)
    try match input_line(chin) with
    | "suivant" when (match code.(unptr 0) with Call(_) -> true | _ -> false) ->
	begin
	  let ad = (unptr 0) + 1 and sp = (unptr 1) in
	  while (unptr 0) <> ad or (unptr 1) <> sp do
	    exec(code.(unptr 0)); addreg(PC,1)
	  done
	end
    | "pas" | "suivant" -> exec(code.(unptr 0)); addreg(PC,1)
    | "goto" ->
	begin
	  let ad = int_of_string(input_line chin) in
	  exec(code.(unptr 0)); addreg(PC,1);
	  while (unptr 0) <> ad do exec(code.(unptr 0)); addreg(PC,1) done
	end
    | "raz"  -> reset()
    | "fin"  -> encore := false
    | _      -> ()
    with HALT -> ()

  done;

  (* arrte wish *)
  fprintf chout "exit\n";
  flush chout;
  let _ = close_process (chin,chout) in ()
;;
#close "unix";;
