(*+--------------------------------------------------------------------------+
  |                                                                          |
  |                       Micro-Caml: typage d'une expression                |
  |                                                                          |
  +--------------------------------------------------------------------------+*)

type mtype = {mutable t:ctype}
and  ctype =
  | Entier
  | Fonction of mtype * mtype
  | Inconnu    of int * int      (* numro, ge *)
  | Quelconque of int;;          (* numro      *)
let tent = {t=Entier};;


                           (*+---------------------+
                             |  Unifie deux types  |
                             +---------------------+*)

exception Errtype of mtype*mtype;;

let rec unifie t1 t2 = if t1 != t2 then match (t1.t,t2.t) with

| (Inconnu(_,a),Inconnu(_,b)) when a < b -> t2.t <- t1.t
| (Inconnu(_),Inconnu(_))                -> t1.t <- t2.t
| (Inconnu(_,a),_) when absent t1 t2 -> rajeunit a t2; t1.t <- t2.t
| (_,Inconnu(_,a)) when absent t2 t1 -> rajeunit a t1; t2.t <- t1.t
| (Entier,Entier)               -> ()
| (Fonction(a,b),Fonction(c,d)) -> unifie a c; unifie b d
| _ -> raise (Errtype(t1,t2))

and absent t t' = match t'.t with
| Fonction(a,b) -> (absent t a) & (absent t b)
| _             -> t.t != t'.t


and rajeunit a t = match t.t with
| Fonction(x,y) -> rajeunit a x; rajeunit a y
| Inconnu(i,b) when b > a -> t.t <- Inconnu(i,a)
| _ -> ()
;;


           (*+-------------------------------------------+
             |    Gnralise les inconnues d'ge >= n    |
             +-------------------------------------------+*)

let rec gnralise n t = match t.t with
| Inconnu(i,a) when a >= n -> t.t <- Quelconque(i)
| Fonction(t1,t2) -> gnralise n t1; gnralise n t2
| _ -> ()
;;


   (*+----------------------------------------------------------+
     |  cre une nouvelle inconnue pour chaque type quelconque  |
     +----------------------------------------------------------+*)

let numinc    = ref 0;;
let new_inc() = incr numinc; {t=Inconnu(!numinc,!numinc)};;


let dupplique t =

  let table = ref [] in

  let rec copy(t) = match t.t with
  | Entier -> tent
  | Fonction(a,b) -> {t=Fonction(copy a, copy b)}
  | Inconnu(_)    -> t
  | Quelconque(i) -> try assoc i !table
	with Not_found -> let t' = new_inc() in table := (i,t') :: !table; t'

  in copy(t)
;;

                                (*+----------+
                                  |  Typage  |
                                  +----------+*)


(* type une expression *)
let rec typ_exp mmoire exp = match exp with

| Const(_) -> tent
| Nom(x)   -> dupplique(assoc x mmoire)

| Fun(x,e) ->
    let tx = new_inc()                   in
    let te = typ_exp ((x,tx)::mmoire) e in
    {t=Fonction(tx,te)}

| Bin(a,_,b) ->
    unifie (typ_exp mmoire a) tent;
    unifie (typ_exp mmoire b) tent;
    tent

| Appl(f,x) ->
    let tx = typ_exp mmoire x in
    let tr = new_inc()         in
    unifie (typ_exp mmoire f) {t=Fonction(tx,tr)};
    tr

| If(c,v,f) ->
    unifie (typ_exp mmoire c) tent;
    let tv = typ_exp mmoire v in
    unifie tv (typ_exp mmoire f);
    tv

| Letin(x,e,f) ->
    let tx = new_inc()         in
    let n  = !numinc          in
    let m  = (x,tx) :: mmoire in
    unifie (typ_exp m e) tx;
    gnralise n tx;
    typ_exp m f
;;


(* type une phrase *)
let mem_type = ref [];;

let typage(texte) = match arbre(mots(texte)) with
| E(e)     -> let t = typ_exp !mem_type e in gnralise 0 t; t
| Def(x,e) -> let tx = new_inc()            in
              let mem = (x,tx) :: !mem_type in
	      unifie tx (typ_exp mem e);
	      gnralise 0 tx;
	      mem_type := mem;
	      tx
| _ -> failwith "cas impossible"
;;

                          (*+-----------------------+
                            |  Impression linaire  |
                            +-----------------------+*)

#open "format";;
let print_type(t) =

  (* conversion inconnue -> lettre *)
  let table = ref [] and lettre = ref(96) in
  let nom(i) =
    try assoc i !table
    with Not_found -> incr lettre; 
                      let l = make_string 1 (char_of_int !lettre) in
		      table := (i,l) :: !table;
		      l
  in

  (* affiche un type *)
  let rec affiche(t) = match t.t with
  | Entier        -> print_string "int"
  | Inconnu(i,_)  -> print_string "_"; print_string(nom i)
  | Quelconque(i) -> print_string (nom i)
  | Fonction(a,b) -> affiche_par a; print_string " -> "; affiche b
  and affiche_par(t) = match t.t with
  | Entier | Inconnu(_) | Quelconque(_) -> affiche(t)
  | Fonction(_) -> print_string "("; affiche t; print_string ")"
  in

  affiche(t)
;;
install_printer "print_type";;
#close "format";;

