#open "tp-defrag";;
load_object "tp-defrag";;

             (*+---------------------------------------+
               |  charge rcursivement l'arborescence  |
               +---------------------------------------+*)


(* pre = rpertoire pre               *)
(* pos  = position du fils dans le pre *)
(* liste = rfrence sur le rsultat    *)
let rec read_arborescence(disque,fat,pre,pos,liste) =

  let fic = ref [] in (* liste des fichiers/s-rep. *)

  (* lit tous les blocs et limine les entres S *)
  let i = ref(pre.(pos).dbut) in
  while !i > 0 do
    let dir = dir_of_secteur(disque.lit_secteur !i) in
    for j=0 to vect_length(dir)-1 do
      let d = dir.(j) in match d.typ with
      |	S -> ()
      |	_ -> fic := d :: !fic
    done;
    i := fat.(!i)
  done;
  let dir = vect_of_list(rev !fic) in (* reconstitue un tableau *)

  (* parcourt les sous-rpertoires *)
  for i=vect_length(dir)-1 downto 0 do
    let d = dir.(i) in if d.typ = D
    then read_arborescence(disque,fat,dir,i,liste)
  done;

  liste := (dir,pre,pos) :: !liste
;;


  (*+------------------------------------------------------------+
    |  constitue la fat dfragmente et la table des mouvements  |
    +------------------------------------------------------------+*)

let rorganise(disque,fat,arb) =

  let f   = make_vect (vect_length fat) 0 (* future fat *)
  and mvt = make_vect disque.ns 0         (* mouvements *)
  in

  (* allocation de n secteurs *)
  let s = ref 0 in
  let alloue(n) =
    for i = !s to !s+n-2 do f.(i) <- i+1 done;
    f.(!s+n-1) <- -1;
    s := !s+n
  in

  alloue(disque.nf); (* fat *)

  (* alloue les rpertoires et les fichiers *)
  let job(dir,pre,pos) =

    (* reloge le rpertoire *)
    pre.(pos) <- {typ=D; nom = pre.(pos).nom; dbut = !s};
    alloue(max 1 ((vect_length(dir)+disque.cd-1)/disque.cd));

    (* reloge les fichiers fils *)
    for i=0 to vect_length(dir)-1 do match dir.(i) with
    | {typ=F; nom=n; dbut=d} ->
	dir.(i) <- {typ=F; nom=n; dbut=(!s)};
	let j = ref(d) in
	while !j > 0 do mvt.(!j) <- !s; incr s; f.(!s-1) <- !s; j := fat.(!j) done;
	f.(!s-1) <- -1
    | _ -> ()
    done

  in do_list job arb;

  (* rsultats *)
  (f,mvt)
;;


              (*+------------------------------------+
                |  effectue les mouvements calculs  |
                +------------------------------------+*)


let dplace(disque,mvt) = for i=0 to vect_length(mvt)-1 do

  let j = ref(i) and k = ref(mvt.(i)) in
  if !k <> 0 & !k <> i then begin

    let s = ref(disque.lit_secteur i) in (* fait un trou *)

    while !k > 0 do               (* dcale la chane issue de i *)
      let t = disque.lit_secteur !k in
      disque.crit_secteur !k !s;
      mvt.(!j) <- 0;
      j := !k;
      k := mvt.(!k);
      s := t
    done

  end

done;;


                   (*+---------------------------+
                     |  Rcrit les rpertoires  |
                     +---------------------------+*)



let write_dir disque (dir,pre,pos) =

  (* complte  un nombre entier de secteurs *)
  let taille = max 1 ((vect_length(dir)+disque.cd-1)/disque.cd) in
  let d = make_vect (taille*disque.cd) {typ=S; nom=""; dbut=0} in
  blit_vect dir 0 d 0 (vect_length dir);

  (* transfre sur disque *)
  let org = pre.(pos).dbut in
  for i=0 to taille-1 do
    disque.crit_secteur (org+i) (DIR(sub_vect d (i*disque.cd) disque.cd))
  done
;;


                       (*+-------------------+
                         |  Dfragmentation  |
                         +-------------------+*)


let defrag(disque) =

  disque.nlect <- 0;
  disque.necr  <- 0;
  let fat  = read_fat(disque) in
  let pre = [|{typ=D; nom=""; dbut=disque.nf}|] in
  let arb  = ref([]) in
  read_arborescence(disque,fat,pre,0,arb);

  let (f,mvt) = rorganise(disque,fat,!arb) in
  dplace(disque,mvt);
  write_fat(disque,f);
  do_list (write_dir disque) !arb
;;


                            (*+---------+
                              |  essai  |
                              +---------+*)


print_string "\n\n***** initialise un disque avec 200 secteurs\n";;
let d = init_disque(200,1);;
affiche(d,SECTEUR);;

print_string "\n\n***** cration de fichiers et rpertoires\n";;
random_fic(d,10,15,10,10);;
random_rep(d,10);;
random_fic(d,10,15,10,10);;
affiche(d,SECTEUR);;

print_string "\n\n***** contrle de cohrence\n";;
vrifie(d);;
enregistre(d,"disque.modele");;

print_string "\n\n***** dfragmentation\n";;
defrag(d);;
printf__printf "ecr = %d, lect = %d\n" d.necr d.nlect;;
affiche(d,SECTEUR);;
print_string "\n\n***** contrle de cohrence\n";;
vrifie(d);;

print_string "\n\n***** comparaison au disque initial\n";;
let e = relit("disque.modele");;
compare(d,e);;

print_string "\n\n***** fin de l'essai\n";;
