                 (* +---------------------------------------------+
                    |  Utilitaires pour le TP de dfragmentation  |
                    +---------------------------------------------+ *)

#open "printf";;

let e_vide = {typ=S; nom=""; dbut=0};;

type arbre = {
mutable a_pos  : int;         (* position dans le rpertoire pre *)
mutable a_org  : int;         (* no 1er secteur *)
mutable a_dir  : entre vect; (* table des fichiers *)
mutable a_fils : arbre list   (* liste des sous-rpertoires *)
};;

exception Disk_full;;

(* fonctions utilitaires *)
let fat_of_secteur = function
| FAT(v) -> v
| _      -> failwith "fat_of_secteur"

and dir_of_secteur = function
| DIR(v) -> v
| _      -> failwith "dir_of_secteur"

and data_of_secteur = function
| DATA(s) -> s
| _       -> failwith "data_of_secteur"

and copy_secteur = function
| FAT(v)      -> FAT(copy_vect v)
| DIR(v)      -> DIR(copy_vect v)
| DATA(n,i,s) -> DATA(n^"",i,s^"")
;;

(* charge la fat en mmoire *)
let read_fat(disque) =
  let n = disque.nlect in
  let v = make_vect (disque.nf*disque.cf) 0 in
  for i=0 to disque.nf-1 do
    blit_vect (fat_of_secteur(disque.lit_secteur i)) 0 v (i*disque.cf) disque.cf
  done;
  disque.nlect <- n+(disque.nf + disque.ms - 1)/disque.ms;
  v
;;

(* recopie la fat sur disque *)
let write_fat(disque,v) =
  let n = disque.necr in
  for i=0 to disque.nf-1 do
    disque.crit_secteur i (FAT(sub_vect v (i*disque.cf) disque.cf))
  done;
  disque.necr <- n+(disque.nf + disque.ms - 1)/disque.ms;
;;

(* dsalloue la chaine de secteurs commenant en i *)
let dsalloue(fat,i) =
  let j = ref(fat.(i)) in
  fat.(i) <- 0;
  while !j > 0 do
    let k = fat.(!j) in
    fat.(!j) <- 0;
    j := k
  done
;;

(* alloue n secteurs dans la fat, renvoie le numro du 1er *)
let alloue(fat,n) =

  let cherche(i) =
    let j = ref(i) in
    while !j < vect_length(fat) & fat.(!j) <> 0 do incr j done;
    if !j < vect_length(fat) then !j else raise Disk_full
  in

  let i0 = cherche(0) in
  let i  = ref(i0) in
  try for k=1 to n-1 do
       let j = cherche(!i+1) in
       fat.(!i) <- j;
       i := j
     done;
     fat.(!i) <- -1;
     i0
  with Disk_full -> dsalloue(fat,i0); raise Disk_full

;;

(* allonge ou racourcit le fichier commenant secteur i, n = nouvelle taille *)
let resize(fat,i,n) =

  let j = ref(i) and k = ref(n) in
  while fat.(!j) > 0 & !k > 1 do j := fat.(!j); decr k done;
  if fat.(!j) > 0 then begin
    dsalloue(fat,fat.(!j));
    fat.(!j) <- -1
  end
  else if !k > 1 then begin
    fat.(!j) <- alloue(fat,!k-1)
  end
;;

(* charge un rpertoire en mmoire  partir du secteur i *)
let read_dir(disque,fat,i) =
  let n = ref 1 and j = ref i in
  while fat.(!j) > 0 do incr n; j := fat.(!j) done;
  let dir = make_vect (!n*disque.cd) e_vide and j = ref i in
  for s=0 to !n-1 do
    try
      blit_vect (dir_of_secteur(disque.lit_secteur !j)) 0 dir (s*disque.cd) disque.cd;
      j := fat.(!j)
    with Failure(s) -> failwith (sprintf "%s secteur %d" s !j)
  done;
  dir
;;

(* charge un fichier en mmoire  partir du secteur i *)
let read_file(disque,fat,i) =
  let i = ref i and l = ref[] in
  while !i > 0 do l := disque.lit_secteur(!i) :: !l; i := fat.(!i) done;
  vect_of_list(rev(!l))
;;

(* charge l'arborescence complte du disque *)
let rec read_arbre(disque,fat,pos,org) =

  if org <= 0 or org >= disque.ns then failwith (sprintf "numro de secteur illgal : %d" org);
  let d = read_dir(disque,fat,org) in
  let fils = ref([]) in
  for j=vect_length(d)-1 downto 0 do match d.(j) with
  | {typ=D; dbut=d} -> fils := (j,d) :: !fils
  | _ -> ()
  done;
  {a_pos  = pos;
   a_org  = org;
   a_dir  = d;
   a_fils = map (fun (p,d) -> read_arbre(disque,fat,p,d)) !fils}
;;

(* rcrit l'arborescence, les secteurs sont dj allous *)
let rec write_arbre(disque,fat,arbre) =

  let i = ref(arbre.a_org) in
  for s=0 to (vect_length arbre.a_dir)/disque.cd - 1 do
    disque.crit_secteur !i (DIR(sub_vect arbre.a_dir (s*disque.cd) disque.cd));
    i := fat.(!i)
  done;

  do_list (fun a -> write_arbre(disque,fat,a)) arbre.a_fils
;;

(* cherche une entre libre dans un rpertoire et l'allonge si ncessaire *)
let find_room(disque,fat,arbre) =

  (* recherche une entre libre *)
  let pos = ref(0) in
  while !pos < vect_length(arbre.a_dir) & arbre.a_dir.(!pos).typ <> S do incr pos done;

  (* s'il n'y a plus de place, allonge le rpertoire *)
  if !pos = vect_length(arbre.a_dir) then begin
    resize(fat,arbre.a_org,vect_length(arbre.a_dir)/disque.cd + 1);
    arbre.a_dir <- concat_vect arbre.a_dir (make_vect disque.cd e_vide)
  end;

  !pos
;;

(* cherche un nom non utilis dans un rpertoire *)
let nouveau_nom(rep) =
  let nom = make_string 5 ` ` in
  let encore = ref true in
  while !encore do
    encore := false;
    for i=0 to 4 do nom.[i] <- char_of_int(97+random__int(26)) done;
    for i=0 to vect_length(rep)-1 do if rep.(i).nom = nom then encore := true done;
  done;
  nom
;;

(* cre un fichier sur disque, et renvoie son index *)
let make_file(disque,fat,arbre,nom,taille) =

  let pos = find_room(disque,fat,arbre) in
  let i   = ref(alloue(fat,taille))     in

  arbre.a_dir.(pos) <- {typ=F; nom=nom; dbut = !i};
  for s=0 to taille - 1 do
    disque.crit_secteur !i (DATA(nom,s,""));
    i := fat.(!i)
  done;

  pos
;;

(* supprime un fichier *)
let remove_file(disque,fat,arbre,pos) =
  let {nom=n; dbut=d} = arbre.a_dir.(pos) in
  arbre.a_dir.(pos) <- {typ=S; nom=n; dbut=d};
  dsalloue(fat,d)
;;

(* cre un rpertoire *)
let make_dir(disque,fat,arbre,nom) =

  let pos = find_room(disque,fat,arbre) in
  let i   = alloue(fat,1)               in
  let d   = make_vect disque.cd e_vide  in
  arbre.a_dir.(pos) <- {typ=D; nom=nom; dbut=i};
  disque.crit_secteur i (DIR(d));
  arbre.a_fils <- {a_pos = pos;
                   a_org = i;
                   a_dir = d;
                   a_fils = []} :: arbre.a_fils
;;

(* supprime un rpertoire et remonte les fichiers associs d'un cran *)
let remove_dir(disque,fat,arbre,pos) =

  let {typ=t; nom=n; dbut=d} = arbre.a_dir.(pos) in
  if t <> D then failwith "pas un rpertoire";

  let (a,reste) =
    let rec supprime(l) = match l with
    | [] -> failwith "rpertoire inexistant"
    | a::suite -> if a.a_pos = pos then (a,suite)
                                   else let (b,reste) = supprime(suite) in (b,a::reste)
    in supprime(arbre.a_fils)
  in

  arbre.a_dir.(pos) <- {typ=S; nom=n; dbut=d};
  arbre.a_fils <- reste;
  do_list (fun b -> b.a_org <- b.a_org + vect_length(arbre.a_dir)) a.a_fils;
  arbre.a_dir <- concat_vect arbre.a_dir a.a_dir;

  let i = ref(arbre.a_org) in
  while fat.(!i) > 0 do i := fat.(!i) done;
  fat.(!i) <- a.a_org
;;

(* formattage d'un disque  ns secteurs, ms bocs *)
let init_disque(ns,ms) =

  let image = make_vect ns (DATA("",0,""))
  and cf = 256
  and cd =  32 in

  let nf = (ns+cf-1)/cf in
  if nf+1 >= ns then raise Disk_full;

  (* initialisation de la fat *)
  let v = make_vect (nf*cf) 0 in
  for i=0  to nf-2 do v.(i) <- i+1 done;      (* sect. Fat    *)
  v.(nf-1) <- -1;                             (* fin Fat      *)
  v.(nf)   <- -1;                             (* sect. Dir    *)
  for i=ns to (nf*cf-1) do v.(i) <- -2 done;  (* sect. inval. *)
  for i=0  to nf-1 do image.(i) <- FAT(sub_vect v (i*cf) cf) done;

  (* initialisation du rpertoire racine *)
  image.(nf) <- DIR(make_vect cd e_vide);

  (* fonctions d'accs *)
  let rec disque = {
    ns=ns; cf=cf; nf=nf; cd=cd; ms=ms; necr = 0; nlect = 0;
    lit_secteur   = (fun i     -> disque.nlect <- disque.nlect+1;
                                  copy_secteur(image.(i)));
    crit_secteur = (fun i s   -> disque.necr  <- disque.necr+1;
                                  image.(i) <- copy_secteur(s));
    lit_secteurs  = (fun (i,n) -> disque.nlect <- disque.nlect + (n+ms-1)/ms;
                                  let l = ref [] in
                                  for j=i+n-1 downto i do
                                    l := copy_secteur(image.(j)) :: !l
                                  done;
                                  !l);
    crit_secteurs= (fun i l   -> let l = ref l and j = ref i in
                                  while !l <> [] do
                                    image.(!j) <- copy_secteur(hd !l);
                                    incr j;
                                    l := tl !l
                                  done;
                                  disque.necr <- disque.necr + (!j - i + ms)/ms)
  }
  in disque
;;

(* simule n fois c crations et e effacements *)
let random_fic(disque,n,c,e,taille) =

  let fat   = read_fat(disque)                    in
  let arbre = read_arbre(disque,fat,-1,disque.nf) in

  (* constitue la table des rpertoires *)
  let l = ref [] in
  let rec job(a) = l := a :: !l; do_list job a.a_fils in job(arbre);
  let rep = vect_of_list(!l) in

  (* constitue la liste des fichiers *)
  let fic = ref [] and l = ref(0) in
  let job(a) = for i=0 to vect_length(a.a_dir)-1 do
                 if a.a_dir.(i).typ=F then begin
                    fic := (a,i) :: !fic;
                    incr l
                 end
               done
  in  do_vect job rep;

  (* simulation de crations-suppressions *)
  for essai=1 to n do

    for cr=1 to c do (* cre c fichiers *)
      try let a = rep.(random__int(vect_length rep)) in
          let n = nouveau_nom(a.a_dir)               in
          let t = 1 + random__int taille             in
          let p = make_file(disque,fat,a,n,t)        in
          fic := (a,p) :: !fic;
          incr l
      with Disk_full -> ()
    done;

    for dl=1 to min e !l do (* en efface e *)
      let j = random__int(!l) in
      let f = ref[] in
      for i = 1 to j do f := hd(!fic) :: !f; fic := tl(!fic) done;
      let (a,p) = hd(!fic) in
      fic := rev(!f) @ tl(!fic);
      decr l;
      remove_file(disque,fat,a,p)
    done

  done; (* for essai *)

  (* mise  jour fat et rep *)
  write_fat(disque,fat);
  write_arbre(disque,fat,arbre)
;;

(* cre n sous-rpertoires *)
let random_rep(disque,n) =

  let fat   = read_fat(disque)                    in
  let arbre = read_arbre(disque,fat,-1,disque.nf) in

  (* constitue la liste des rpertoires *)
  let rep = ref [] and l = ref 0 in
  let rec job(a) = rep := a :: !rep; incr l; do_list job a.a_fils in job(arbre);

  (* crations alatoires *)
  for essai=1 to n do

    let f = ref(!rep) in
    for i=1 to random__int(!l) do f := tl(!f) done;
    let a = hd(!f)                   in
    let n = nouveau_nom(a.a_dir)     in
    make_dir(disque,fat,a,n);
    rep := hd(a.a_fils) :: !rep;
    incr l

  done;

  (* mise  jour fat et rep *)
  write_fat(disque,fat);
  write_arbre(disque,fat,arbre)
;;

(* affichage semi-graphique de la Fat *)
let print_fat(disque,fat) =

  let largeur = 60 in
  let nlig    = (vect_length(fat)+largeur-1)/largeur in
  print_string "FAT";
  for i=0 to vect_length(fat)-1 do
    if i mod largeur = 0 then print_newline();
    print_string(match fat.(i) with
    | 0   -> "."
    | -2  -> " "
    | j when j < -2 -> "?"
    | _ -> match disque.lit_secteur(i) with
           | FAT(_)  -> "F"
           | DIR(_)  -> "D"
           | DATA(_) -> "x")
  done;
  print_newline()
;;

(* liste le fichier commenant secteur i *)
let print_file(disque,fat,i,champ) =
  let i = ref(i) in
  while !i > 0 do
    let (n,num,s) = data_of_secteur(disque.lit_secteur !i) in
    begin match champ with
    | NOM     -> fprintf stdout "%s " n
    | NUMERO  -> fprintf stdout "%d " num
    | TEXTE   -> fprintf stdout "%s " s
    | SECTEUR -> fprintf stdout "%d " !i
    end;
    i := fat.(!i)
  done;
  print_newline()
;;

(* liste les secteurs d'un rpertoire *)
let print_dir(fat,i) =
  let i = ref(i) in
  while !i > 0 do fprintf stdout "%d " !i; i := fat.(!i) done;
  print_newline()
;;


(* liste rcursivement l'arborescence *)
let rec print_arbre(disque,fat,nom,arbre,champ) =

  fprintf stdout "\n%s\n" (if nom = "" then "\\" else nom);

  for i=0 to vect_length(arbre.a_dir)-1 do match arbre.a_dir.(i) with
  | {typ=F; nom=n; dbut=d} -> fprintf stdout "%s\t" n; print_file(disque,fat,d,champ)
  | {typ=D; nom=n; dbut=d} when champ = SECTEUR -> fprintf stdout "%s\t" n; print_dir(fat,d)
  | {typ=D; nom=n; dbut=d}                      -> fprintf stdout "%s\tDIR\n" n
  | {typ=S} -> ()
  done;

  do_list (fun a -> print_arbre(disque,fat,nom ^ "\\" ^ arbre.a_dir.(a.a_pos).nom,a,champ)) arbre.a_fils
;;

(* liste tout le disque *)
let affiche(disque,champ) =
  let (l,e) = (disque.nlect,disque.necr) in
  let fat   = read_fat(disque) in
  let arbre = read_arbre(disque,fat,-1,disque.nf) in
  print_fat(disque,fat);
  print_arbre(disque,fat,"",arbre,champ);
  flush stdout;
  disque.nlect <- l; disque.necr <- e
;;

(* vrifie la cohrence du disque *)
let vrifie(disque) =

  let (l,e) = (disque.nlect,disque.necr) in

  let err_fat(i)    = failwith (sprintf "Fat incorrecte, secteur %d" i)
  and err_tte(i)   = failwith (sprintf "dbut de fichier en mileu de chaine, secteur %d" i)
  and err_all(n,n') = failwith (sprintf "fichier non allou : %s\\%s" n n')
  and err_ill(i)    = failwith (sprintf "numro de secteur illgal : %d" i)
  and err_dbl(i)    = failwith (sprintf "double allocation, secteur : %d" i)
  and err_lost(i)   = failwith (sprintf "chaine perdue, secteur : %d" i)
  and err_cont(i)   = failwith (sprintf "donnes incorrectes, secteur : %d" i)
  in

  let fat = read_fat(disque)          in (* liens avant         *)
  let f   = make_vect disque.ns (-1)  in (* liens arrire       *)
  let vu  = make_vect disque.ns false in (* dbut de chaine vue *)

  (*--- contrle de la Fat ---*)

  (* cherche les prdecesseurs *)
  for i=0 to disque.ns-1 do let j = fat.(i) in
    if j < -1 then err_fat(i)
    else if j = -1 then ()
    else if j =  0 then (if f.(i) = -1 then f.(i) <- 0 else err_fat(i))
    else if j < disque.ns then (if f.(j) = -1 then f.(j) <- i else err_fat(i))
    else err_fat(i)
  done;

  (* partie Fat *)
  for i=0 to disque.nf-2 do if fat.(i) <> i+1 then err_fat(i) done;
  if fat.(disque.nf-1) <> -1 then err_fat(disque.nf-1);

  (* rpertoire racine *)
  if fat.(disque.nf) = 0 then err_fat(disque.nf);
  if f.(disque.nf) <> -1 then err_tte(disque.nf);

  (*--- contrle de l'arborescence *)
  let arbre = read_arbre(disque,fat,-1,disque.nf) in

  let rec job(a,nom) =
    for i=0 to vect_length(a.a_dir)-1 do match a.a_dir.(i) with
    | {typ=S} -> ()
    | {typ=D; nom=n; dbut=d} ->
        if d <= 0 or d >= disque.ns then err_ill(d);
        if f.(d) = 0   then err_all(nom,n);
        if f.(d) <> -1 then err_tte(d);
        if vu.(d)      then err_dbl(d);
        vu.(d) <- true;
        let d = ref(d) in
        while !d > 0 do
          try
            let _ = dir_of_secteur(disque.lit_secteur !d) in
            d := fat.(!d);
          with Failure "dir_of_secteur" -> err_cont(!d)
        done
    | {typ=F; nom=n; dbut=d} ->
        if d <= 0 or d >= disque.ns then err_ill(d);
        if f.(d) = 0   then err_all(nom,n);
        if f.(d) <> -1 then err_tte(d);
        if vu.(d)      then err_dbl(d);
        vu.(d) <- true;
        let d = ref(d) and s = ref(0) in
        while !d > 0 do
          try
            let (n',s',_) = data_of_secteur(disque.lit_secteur !d) in
            if n <> n' or !s <> s' then err_cont(!d);
            d := fat.(!d);
            incr s
          with Failure "data_of_secteur" -> err_cont(!d)
        done
    done;
    do_list (fun b -> job(b,nom ^ "\\" ^ a.a_dir.(b.a_pos).nom)) a.a_fils
  in job(arbre,"");

  (*--- chaines perdues ---*)
  for i=disque.nf+1 to disque.ns-1 do if f.(i) = -1 & not(vu.(i)) then err_lost(i) done;

  (*--- restaure les compteurs ---*)
  disque.nlect <- l; disque.necr <- e
;;


(* sauvegarde/relecture d'un disque *)
let enregistre(disque,nom) =

  let (l,e) = (disque.nlect,disque.necr) in
  let canal = open_out(nom) in

  output_value canal (disque.ns, disque.cf, disque.nf, disque.cd, disque.ms);
  for i=0 to disque.ns-1 do output_value canal (disque.lit_secteur i) done;
  close_out canal;

  disque.nlect <- l; disque.necr <- e

;;

let relit(nom) =

  let canal = open_in(nom) in
  let (ns,cf,nf,cd,ms) = input_value(canal) in
  let d = init_disque(ns,ms) in
  for i=0 to ns-1 do d.crit_secteur i (input_value canal) done;
  close_in(canal);
  d.necr <- 0;
  d
;;

(* dit si deux disques contiennent les mmes infos *)
let compare(disque1,disque2) =

  let fat1 = read_fat(disque1) and fat2 = read_fat(disque2) in

  (* trie un rpertoire *)
  let tri(dir) = sort__sort avant (list_of_vect dir)
    where avant e1 e2 = match (e1,e2) with
    | _ when e1.typ = e2.typ -> e1.nom <= e2.nom
    | {typ=D},_ -> true
    | _,{typ=D} -> false
    | {typ=F},_ -> true
    | _         -> false
  in

  (* cas d'erreur *)
  let err_abs(x,nom) = failwith (sprintf "absent sur le disque %d : %s" x nom)
  and err_typ(nom)   = failwith (sprintf "types diffrents : %s" nom)
  and err_file(nom)  = failwith (sprintf "contenus diffrents : %s" nom)
  in

  (* compare rcursivement deux arborescences *)
  let rec compare(s1,s2,nom) =

    (* compare deux fichiers *)
    let rec cmp_fic(e1,e2,nom) =
      let i1 = ref(e1.dbut) and i2 = ref(e2.dbut) in
      while !i1 > 0 & !i2 > 0 do
        if disque1.lit_secteur !i1 <>
           disque2.lit_secteur !i2 then err_file(nom);
        i1 := fat1.(!i1);
        i2 := fat2.(!i2)
      done;
      if !i1 >= 0 or !i2 >= 0 then err_file(nom)
    in

    (* compare deux listes d'entres *)
    let rec cmp_rep(l1,l2) =

      let e1 = (match l1 with ([] | {typ=S}::_) -> e_vide | _ -> hd(l1))
      and e2 = (match l2 with ([] | {typ=S}::_) -> e_vide | _ -> hd(l2)) in
      let n  = nom ^ "\\" ^ e1.nom in

      if      e1.nom < e2.nom then err_abs(2,n)
      else if e2.nom < e1.nom then err_abs(1,nom ^ "\\" ^ e2.nom)

      else match (e1.typ,e2.typ) with
      | S,S -> ()
      | S,_ -> err_abs( 1,n)
      | _,S -> err_abs( 2,n)
      | F,F -> cmp_fic(e1,e2,n);             cmp_rep (tl l1, tl l2)
      | D,D -> compare(e1.dbut,e2.dbut,n); cmp_rep (tl l1, tl l2)
      | _   -> err_typ(n)

    in
    cmp_rep(tri(read_dir(disque1,fat1,s1)), tri(read_dir(disque2,fat2,s2)))

  in compare(disque1.nf,disque2.nf,"")

;;
