                    (* +---------------------------------+
                       |  Dterminisation d'un automate  |
                       +---------------------------------+ *)


include "tp-automate.ml";;


                             (* +---------------+
                                |  Utilitaires  |
                                +---------------+ *)


(* fusion sans rptition de deux listes tries *)
let rec fusion l l' = match 0 with
|_ when l=[]  -> l'
|_ when l'=[] -> l
|_ when hd(l) < hd(l') -> hd(l)  :: fusion (tl l) l'
|_ when hd(l) > hd(l') -> hd(l') :: fusion l (tl l')
|_ -> hd(l) :: fusion (tl l) (tl l')
;;
#infix "fusion";;

(* appartenance  une liste trie *)
let rec appartient x l = match 0 with
|_ when l=[] -> false
|_ when hd(l) = x -> true
|_ when hd(l) < x -> appartient x (tl l)
|_ -> false
;;
#infix "appartient";;


(* cherche si deux listes tries ont un lment en commun *)
let rec inter l l' = match 0 with
|_ when l=[]  -> false
|_ when l'=[] -> false
|_ when hd(l) < hd(l') -> inter (tl l) l'
|_ when hd(l) > hd(l') -> inter l (tl l')
|_ -> true
;;
#infix "inter";;


                        (* +-------------------------+
                           |  table de mmorisation  |
                           +-------------------------+ *)

type 'a table = { taille : unit -> int; numro  : 'a -> int; lment : int -> 'a };;

(*
 * "t.taille()" renvoie le nombre d'lments enregistrs dans la table.
 * Ces lments sont numrots  partir de 0.
 *
 * "t.numro(x)" enregistre "x" s'il n'est pas dj dans la table et renvoie
 * dans tous les cas le numro d'ordre de "x".
 *
 * "t.lment(n)" renvoie l'lment de numro d'ordre "n".
 *)

let new_table() =

  let rec lm(n,m) =
    if n=0 then hd(m) else lm(n-1,tl(m))

  and num(x,m) =
    if m=[] then (0,[x])
    else if hd(m) = x then (0,m)
    else let (n,m') = num(x,tl(m)) in (n+1, hd(m)::m')
  in

  let mmoire = ref([]) in
  { taille = (fun () -> list_length(!mmoire));
    numro = (fun x -> let (n,m) = num(x,!mmoire) in mmoire := m; n);
    lment= (fun n -> lm(n,!mmoire)) }
;;

(* Rmq : implmentation peu efficace, utiliser une table de hachage *)



             (* +----------------------------------------------+
                |  Calcule les cltures par transitions vides  |
                +----------------------------------------------+ *)

(* renvoie un vecteur v tel que v.(i) = cloture(i) *)

let cltures {nb_tats=n; t_vide=liste} =

  let v = make_vect n [] in
  for i=0 to n-1 do v.(i) <- [i] done;

  begin
    do_list traitement liste
    where traitement(x,y) =
      for z=0 to n-1 do
       if (x appartient v.(z)) & not(y appartient v.(z)) then v.(z) <- v.(z) fusion v.(y)
      done
  end;
  v
;;

                  (* +------------------------------------+
                     |  Successeur d'un ensemble d'tats  |
                     +------------------------------------+ *)


let succ (clture,{t_lettre=liste}) lettre ens =
  let res = ref [] in
  do_list 
    (fun (x,a,y) -> if (a = lettre) & (x appartient ens)
                    then res := !res fusion clture.(y))
    liste;
  !res
;; 

                           (* +-------------------+
                              |  Dterminisation  |
                              +-------------------+ *)

let dterminise(a) =

  let cl   = cltures(a) in           (* cltures                    *)
  let succ = succ(cl,a)  in           (* abrviation pour succ       *)
  let t    = new_table() in           (* table des ensembles d'tats *)
  let tr   = ref([])     in           (* liste des transitions       *)

  (* calcule l'ensemble des tats de dpart *)
  let d = ref [] in
  do_list (fun x -> d := !d fusion cl.(x)) a.dpart;
  let _ = t.numro(!d) in

  let n = ref(0) in                   (* numro de l'ensemble tudi *)
  while !n < t.taille() do let tat = t.lment(!n) in

    (* gnre les successeurs de "tat" et enregistre les transitions *)
    do_list
      (fun u -> let e = succ u tat in
                let p = t.numro(e) in
                tr := (!n, u, p) :: !tr)
      a.alphabet;

    incr n;
  done;

  (* recherche les nouveaux tats finals *)
  let finals = ref [] in
  for i=t.taille()-1 downto 0 do 
    if t.lment(i) inter a.finals then finals := i :: !finals
  done;

  (* assemble le nouvel automate *)
  { alphabet = a.alphabet;
    nb_tats = t.taille();
    dpart   = [0];
    finals   = !finals;
    t_vide   = [];
    t_lettre = !tr }
;;

                  (* +------------------------------------+
                     |  Expression rgulire -> automate  |
                     +------------------------------------+ *)

(* exp = expression  transformer *)
let thompson(exp) =

  let alphabet = ref []  (* alphabet en cours de construction *)
  and t_vide   = ref []  (* transitions vides     *)
  and t_lettre = ref []  (* tr. sur lettre        *)
  and n = ref(0)         (* 1er numro disponible *)
  in

  (*
   * traitement du sous-arbre "e". Met  jour "alphabet", "t_vide", "t_lettre"
   * et "n" et renvoie les numros des tats initial et final du sous-automate
   * constitu.
   *)
  let rec traitement(e) = match e with

  | Mot(m) ->
     for i=0 to vect_length(m)-1 do
       alphabet := !alphabet fusion [m.(i)];
       t_lettre := (!n+i, m.(i), !n+i+1) :: !t_lettre
     done;
     let res = (!n, !n + vect_length(m)) in n := !n + vect_length(m) + 1; res

  | Somme(e1,e2) ->
     let (d1,f1) = traitement(e1)
     and (d2,f2) = traitement(e2) in
     t_vide := (!n,d1) :: (!n,d2) :: (f1,!n+1) :: (f2,!n+1) :: !t_vide;
     let res = (!n, !n + 1) in n := !n + 2; res

  | Concat(e1,e2) ->
     let (d1,f1) = traitement(e1)
     and (d2,f2) = traitement(e2) in
     t_vide := (!n,d1) :: (f1,d2) :: (f2,!n+1) :: !t_vide;
     let res = (!n, !n + 1) in n := !n + 2; res

  | Etoile(e1) ->
     let (d1,f1) = traitement(e1) in
     t_vide := (!n,d1) :: (f1,!n) :: !t_vide;
     let res = (!n, !n) in n := !n + 1; res

  in
  let (d,f) = traitement(exp) in
  {alphabet = !alphabet;
   nb_tats = !n;
   dpart   = [d];
   finals   = [f];
   t_vide   = !t_vide;
   t_lettre = !t_lettre}

;;


                            (*+---------+
                              |  essai  |
                              +---------+*)

let e = expression "(a+b)*(a+c)*";;
let a = thompson(e);;
let b = dterminise(a);;
