     (* +---------------------------------------------------------------+
        |  Utilitaires sur les expressions rgulires et les automates  |
        +---------------------------------------------------------------+ *)

#open "genlex";;
#open "arbres";;
load_object "arbres";;

                           (* +------------------+
                              |  Types utiliss  |
                              +------------------+ *)


type 'a automate =      		(* 'a est le type des lettres *)
{ alphabet : 'a list;		     
  nb_tats : int;       		(* les tats sont numrots de 0  n-1 *)
  dpart   : int list;			(* liste des tats de dpart           *)
  finals   : int list;			(* liste des tats finals              *)
  t_vide   : (int * int) list;          (* tr. vides (origine,extrmit)       *)
  t_lettre : (int * 'a * int) list };;  (* tr. sur lettre (org., lettre, ext.) *)

type 'a expression =    (* expression rgulire sur l'alphabet 'a *)
| Mot    of 'a vect
| Somme  of ('a expression) * ('a expression)
| Concat of ('a expression) * ('a expression)
| Etoile of ('a expression)
;;


                    (* +---------------------------------+
                       |  conversion chane <-> vecteur  |
                       +---------------------------------+ *)


let vect_of_string(s) =
  let l = string_length(s) in
  let v = make_vect l ` `  in
  for i=0 to l-1 do v.(i) <- s.[i] done;
  v

and string_of_vect(v) =
  let l = vect_length(v) in
  let s = make_string l ` `  in
  for i=0 to l-1 do s.[i] <- v.(i) done;
  s
;;

let expression =
  let lexer = make_lexer ["+"; "*"; "."; "("; ")";]
  and syntax_of_lexme = function
      | Kwd "+" -> Infixe(1,fun x y -> Somme(x,y))
      | Kwd "." -> Infixe(2,fun x y -> Concat(x,y))
      | Kwd "*" -> Postfixe(fun x -> Etoile(x))
      | Kwd "(" -> Lpar
      | Kwd ")" -> Rpar
      | Ident(i)-> Valeur(Mot(vect_of_string i))
      | _       -> raise Parse_error
  in let eval = analyseur (fun () -> (2,fun x y -> Concat(x,y)))
                          syntax_of_lexme lexer
  in fun s -> try eval s
  with Erreur(exn,i) -> print_erreur s i; raise exn
;;


let rec g_arbre_of_exp = function
| Mot(m) -> G_noeud(string_of_vect m,[])
| Somme(u,v)  -> G_noeud("+", map g_arbre_of_exp [u;v])
| Concat(u,v) -> G_noeud(".", map g_arbre_of_exp [u;v])
| Etoile(u)   -> G_noeud("*", map g_arbre_of_exp [u])
;;

let print_expression(e) = print_string_g_arbre(g_arbre_of_exp e);;
install_printer "print_expression";;

#close "genlex";;
#close "arbres";;
