(*+-------------------------------------------------------------------------+
  |                                                                         |
  |                       Calcul de dveloppements limits                  |
  |                                                                         |
  +-------------------------------------------------------------------------+*)

include "tp-dl.ml";;



                        (* +-------------------------+
                           |  Oprations sur les dl  |
                           +-------------------------+ *)


let addition(a,b,op) = (* a,b = dl; op = +. ou -. *)
  let v = min (valuation a) (valuation b)
  and p = min (prcision a) (prcision b) in
  let dl = cre_dl(v,p) in
  for k = v to p do store(dl,k, op (coeff(a,k)) (coeff(b,k))) done;
  normalise(dl)
;;


let add_cste(a,c) = (* a=dl, c=nombre *)
  let b = cre_dl(0,prcision(a)) in
  store(b,0,c);
  addition(a,b,prefix +.)
;;

let multiplication(a,b,n) = (* a,b = dl, tronque le produit  l'ordre n *)
  let v = valuation(a) + valuation(b) in
  let dl = cre_dl(v,n) in
  for k = v to n do
    let c = ref(zro) in
    for i = valuation(a) to k-valuation(b) do
      c := !c +. coeff(a,i) *. coeff(b,k-i)
    done;
    store(dl,k,!c)
  done;
  dl
;;


let division(a,b,n) = (* a,b = dl, tronque le quotient  l'ordre n *)

  let v     = valuation(a) - valuation(b) in
  let dl    = cre_dl(v,n)
  and reste = {k0=a.k0; coeff = copy_vect a.coeff}
  and b0    = coeff(b,valuation(b))
  and vb    = valuation(b)
  in

  for k=v to n do
    let c = coeff(reste,k+vb) /. b0 in
    for i=k+vb to n+vb do
      store(reste,i, coeff(reste,i) -. c *. coeff(b, i-k))
    done;
    store(dl,k,c)
  done;
  dl
;;


let compose(a,b,n) = (* a,b = dl, tronque la compose  l'ordre n *)

  let va = valuation(a) and vb = valuation(b) in
  if vb <= 0 then failwith "composition impossible";
  let v  = va * vb in

  (* calcule la compose par l'algorithme de Horner *)
  let dl = ref(cre_dl(0,0)) in
  for k=(n-v)/vb downto 0 do
    dl := multiplication(!dl,b,n-v-k*vb);
    dl := add_cste(!dl,coeff(a,k+va))
  done;

  (* multiplie par b^va *)
  for k = 1 to va do
    dl := multiplication(!dl,b,n-v+k*vb)
  done;
  for k = -1 downto va do
    dl := division(!dl,b,n-v+k*vb)
  done;

  !dl  (* rsultat *)

;;


let intgre(dl,c) = (* intgre et ajoute la constante c *)

  if valuation(dl) < 0 then failwith "intgrale divergente";
  let d = {k0=dl.k0+1; coeff=copy_vect dl.coeff} in
  for k = valuation(d) to prcision(d) do
    store(d,k, coeff(d,k) /. nombre_of_int(k))
  done;
  add_cste(d,c)

;;

                        (* +-------------------------+
                           |  Dveloppements usuels  |
                           +-------------------------+ *)


(* fonction trigonomtrique ou hyperbolique. Les numrateurs sont *)
(* successivement a, b, c, d, a, b, ...                           *)
let dl_trig (a,b,c,d,n) =
 
  let dl = cre_dl(0,n)
  and roue = ref(a,b,c,d)
  and den  = ref(un)
  in

  for k=0 to n do
    let (a,b,c,d) = !roue in
    if k >= 2 then den := !den *. nombre_of_int(k);
    store(dl,k, a /. !den);
    roue := (b,c,d,a)
  done;
  normalise(dl)
;;
    
let dl_exp(x,n) = let a=exp(x) in dl_trig(a,a,a,a,n)
and dl_cos(x,n) = let a=cos(x) in let b = sin(x) in dl_trig(a,opp(b),opp(a),b,n)
and dl_sin(x,n) = let a=cos(x) in let b = sin(x) in dl_trig(b,a,opp(b),opp(a),n)
and dl_ch (x,n) = let a=ch(x)  in let b = sh(x)  in dl_trig(a,b,a,b,n)
and dl_sh (x,n) = let a=ch(x)  in let b = sh(x)  in dl_trig(b,a,b,a,n)
;;

let dl_tan(a,n) = division(dl_sin(a,n+1), dl_cos(a,n+2), n)
and dl_cot(a,n) = division(dl_cos(a,n+1), dl_sin(a,n+2), n)
and dl_th (a,n) = division(dl_sh(a,n),    dl_ch(a,n),    n)
;;

let dl_puiss(alpha,a,n) = (* (a+h)^alpha  l'ordre n, a <> 0 *)

  let dl = cre_dl(0,n)
  and c  = ref(a **. alpha) in

  for k=0 to n do
    store(dl,k,!c);
    let k' = nombre_of_int(k) in
    c := (alpha -. k') *. !c /. (k' +. un) /. a;
  done;
  dl

;;

let dl_log(a,n) = intgre(dl_puiss(opp(un),a,n-1), log(a));;

let usuel(nom) = match nom with
| "exp" -> dl_exp
| "ch"  -> dl_ch
| "sh"  -> dl_sh
| "cos" -> dl_cos
| "sin" -> dl_sin
| "tan" -> dl_tan
| "cot" -> dl_cot
| "th"  -> dl_th
| "log" -> dl_log
| _     -> failwith "fonction inconnue"
;;

(*  faire : arcs, args *)


                     (* +-------------------------------+
                        |  Dveloppement d'une formule  |
                        +-------------------------------+ *)

let maxessais = 10;; (* nb d'essais pour la multiplication et la division *)

let rec dveloppe(f,a,n) = match f with

| Const(c) ->
   let dl = cre_dl(0,n) in
   if n >= 0 then store(dl,0,c);
   normalise(dl)

| X ->
   let dl = cre_dl(0,n) in
   if n >= 0 then store(dl,0,a);
   if n >= 1 then store(dl,1,un);
   normalise(dl)

| Bin(Plus, g,h) -> addition(dveloppe(g,a,n), dveloppe(h,a,n), prefix +.)
| Bin(Moins,g,h) -> addition(dveloppe(g,a,n), dveloppe(h,a,n), prefix -.)

| Bin(Mult, g,h) -> let (dg,dh) = dev_mult(g,h,a,n, 1) in multiplication(dg,dh,n)
| Bin(Div, g,h)  -> let (dg,dh) = dev_mult(g,h,a,n,-1) in division(dg,dh,n)

| Fct(nom,g) ->
   let dg = dveloppe(g,a,n) in
   if  valuation(dg) < 0 then failwith "composition impossible";
   let a' = coeff(dg,0) in
   let df = usuel(nom) (a',n) in
   let h  = Bin(Moins, g, Const(a'))                  (* infiniment petit *)
   and dh = add_cste(dg, opp(a')) in                  (* DL de h en a     *)
   let u  = dev_comp(valuation(df), h , dh , a, n) in (* prolonge le DL de h si ncessaire *)
   compose(df,u,n)

| Bin(Puiss, g, h) -> dveloppe( Fct("exp", Bin(Mult, h, Fct("log", g))), a, n)
   (* Rmq: il faudrait en fait mettre en facteur le terme de tte de g *)
   (* et appliquer la formule du binme. Ceci permettrait de grer le  *)
   (* cas o g est de valuation non nulle, mais le dtail des opra-   *)
   (* tions tient de l'usine  gaz ...                                 *)


(* dveloppe g et h suffisament loin pour pouvoir calculer g*h ou g/h *)
(* signe = 1 pour g*h et signe = -1 pour g/h                          *)
and dev_mult(g,h,a,n,signe) =

   let dg = ref(dveloppe(g,a,n)) in
   let dh = ref(dveloppe(h,a,n)) in
   let essai = ref(1)
   and fini  = ref(false)
   in
   while (!essai < maxessais) & (not !fini) do
     fini := true;
     let pg = n - signe*valuation(!dh)
     and ph = n - valuation(!dg) + (1-signe)*valuation(!dh) in
     if prcision(!dg) < pg then (dg := dveloppe(g,a,pg); fini := false);
     if prcision(!dh) < ph then (dh := dveloppe(h,a,ph); fini := false);
     incr(essai)
   done;
   if !fini then (!dg,!dh) else failwith "prcision inaccessible"


(* dveloppe g suffisament loin pour pouvoir calculer f o g *)
(* vf = valuation de f, dg = dveloppement dj connu de g  *)
and dev_comp(vf,g,dg,a,n) =

  let essai = ref(1)
  and fini  = ref(false)
  and dg    = ref(dg)
  in
  while (!essai <= maxessais) & (not !fini) do
    fini := true;
    let p = n - 2*vf*valuation(!dg) in
    if prcision(!dg) < p then begin
      dg := dveloppe(g,a,p);
      fini := false
    end;
    incr essai
  done;
  if !fini then !dg else failwith "prcision inacessible"

;;

(* ouff !!! *)

(* essai *)
let f = value "sin(tan(x))-tan(sin(x))"
and g = value "sh(th(x))-th(sh(x))";;
(* trace "dveloppe";; *)
dveloppe(Bin(Div,f,g),zro,3);;
