(*+-------------------------------------------------------------------------+
  |                                                                         |
  |                       Calcul de dveloppements limits                  |
  |                                                                         |
  |                           bibliothque utilitaire                       |
  |                                                                         |
  +-------------------------------------------------------------------------+*)

(* M. Quercia, le 08/12/96 *)

            (* +-------------------------------------------------+
               |  Interface avec les bibliothques float et num  |
               +-------------------------------------------------+ *)


#open "num";;

type nombre = N of num | F of float;;
let calculs_exacts = ref true;;

let nombre_of_int(x)   = N(num_of_int x)
and nombre_of_float(x) = F(x);;

let zro = nombre_of_int(0)
and un   = nombre_of_int(1)
and deux = nombre_of_int(2)
and pi   = nombre_of_float(mult_float 4.0 (float__atan 1.0))
and e    = nombre_of_float(float__exp 1.0)
;;


(* comparaison approximative *)
let eps = 1.0e-8;;
let eq_float x y = abs_float(sub_float y x) < eps;;

(* vrifie qu'on peut faire des calculs approchs et sinon dclenche une erreur *)
let approx() = match !calculs_exacts with
| true  -> failwith "calcul approch"
| false -> ()
;;

(*
 * Applique une fonction parmi 2  un ou deux arguments de type nombre
 * La fonction applique est celle correspondant au type commun du ou des
 * arguments, les rationnels sont convertis en rels au besoin.
 *)

let app1 opn opf x = match x with
| N(a) -> N(opn a)
| F(a) -> approx(); F(opf a)

and app2 opn opf x y = match (x,y) with
| N(a), N(b) -> N(opn a b)
| N(a), F(b) -> approx(); F(opf (float_of_num a) b)
| F(a), N(b) -> approx(); F(opf a (float_of_num b))
| F(a), F(b) -> approx(); F(opf a b)

and app2_gen opn opf x y = match (x,y) with
| N(a), N(b) -> opn a b
| N(a), F(b) -> approx(); opf (float_of_num a) b
| F(a), N(b) -> approx(); opf a (float_of_num b)
| F(a), F(b) -> approx(); opf a b
;;

(* oprations prdfinies *)
let prefix +.  = app2 add_num    add_float
and prefix -.  = app2 sub_num    sub_float
and prefix *.  = app2 mult_num   mult_float
and prefix /.  = app2 div_num    div_float
and opp        = app1 minus_num  minus_float
;;

let prefix =.  = app2_gen eq_num eq_float
and prefix <.  = app2_gen lt_num lt_float
and prefix <=. = app2_gen le_num le_float
and prefix >.  = app2_gen gt_num gt_float
and prefix >=. = app2_gen ge_num ge_float
;;
let prefix <>. x y = not(x =. y);;

let is_integer(x) = match x with
| N(a) -> is_integer_num(a)
| F(a) -> false
and is_rationnal(x) = match x with
| N(a) -> true
| F(a) -> false
;;

let floor(x) = match x with
| N(a) -> (match floor_num(a) with Int(b) -> b | _ -> failwith "floor")
| F(a) when ge_float a 0. -> int_of_float(a)
| F(a) -> let b = 1 - int_of_float(a) in
          let c = int_of_float(add_float a (float_of_int b)) in c - b
and ceil(x) = match x with
| N(a) -> (match ceiling_num(a) with Int(b) -> b | _ -> failwith "ceil")
| F(a) when le_float a 0. -> int_of_float(a)
| F(a) -> let b = 1 + int_of_float(a) in
          let c = int_of_float(sub_float a (float_of_int b)) in c + b
;;

let prefix **. x y = match (x,y) with
| N(a), N(b) -> if a = Int(1) then un
                else if is_integer_num(b) then N(power_num a b)
                else (approx(); F(power (float_of_num a) (float_of_num b)))
| N(a), F(b) -> approx(); F(power (float_of_num a) b)
| F(a), N(b) -> approx(); F(power a (float_of_num b))
| F(a), F(b) -> approx(); F(power a b)
;;


let string_of_nombre(x) = match x with
| N(a) -> string_of_num(a)
| F(a) -> string_of_float(a)
;;
let print_nombre(x) = format__print_string(string_of_nombre(x));;



(********************** fonctions usuelles *************************)

(* applique opf  x en le convertissant en flottant sauf si x = x0 *)
(* auquel cas le rsultat est y0.                                  *)
let usuelle x0 y0 opf x =
  if x =. x0 then y0    (* Pb: risque de confusion en cas de calcul approche *)
  else match x with
  | N(a) -> approx(); F(opf(float_of_num a))
  | F(a) -> approx(); F(opf a)
;;

let exp  = usuelle  zro  un    float__exp
and cos  = usuelle  zro  un    float__cos
and sin  = usuelle  zro  zro  float__sin
and asin = usuelle  zro  zro  float__asin
and acos = usuelle  un    zro  float__acos
and atan = usuelle  zro  zro  float__atan
and log  = usuelle  un    zro  float__log
;;

let sqrt(x) = x **. (un/. deux)
;;

let ch(x)  = (exp(x) +. exp(opp x)) /. deux
and sh(x)  = (exp(x) -. exp(opp x)) /. deux
and ach(x) = log(x +. sqrt(x*.x -. un))
and ash(x) = log(x +. sqrt(x*.x +. un))
and ath(x) = log((un +. x)/.(un -. x)) /. deux
;;

let tan(x) = sin(x) /. cos(x)
and cot(x) = cos(x) /. sin(x)
and th(x)  = sh(x)  /. ch(x)
;;

#close "num";;



             (* +-----------------------------------------------+
                |  Primitives d'accs aux coefficients d'un DL  |
                +-----------------------------------------------+ *)


let cre_dl(v,p) =
  if p < v then {k0=p+1; coeff=[||]}
  else {k0 = v; coeff = make_vect (p-v+1) zro}
;;


let coeff(dl,k) =
  if k < dl.k0 then zro
  else if k < dl.k0 + vect_length(dl.coeff) then dl.coeff.(k-dl.k0)
  else failwith "coefficient indtermin"
;;

let store(dl,k,c) =
  if (k < dl.k0) or (k-dl.k0 >= vect_length(dl.coeff))
  then failwith "exposant invalide"
  else dl.coeff.(k-dl.k0) <- c
;;

let valuation(dl) = dl.k0
and prcision(dl) = dl.k0 + vect_length(dl.coeff) - 1
;;



(* limine les premiers coefficients nuls *)
let normalise(dl) =
  let k = ref(0) in
  while (!k < vect_length(dl.coeff)) & (dl.coeff.(!k) =. zro)
  do incr k done;
  {k0 = dl.k0 + !k; coeff = sub_vect dl.coeff !k (vect_length dl.coeff - !k)}
;;


(* tronque  l'ordre n *)
let tronque(dl,n) =

  if prcision(dl) < n  then failwith "coefficient indtermin";
  if valuation(dl) <= n then {k0 = dl.k0; coeff = sub_vect dl.coeff 0 (n - dl.k0 + 1)}
                        else {k0 = n+1; coeff=[||]}

;;

                         (* +----------------------+
                            |  Impression d'un DL  |
                            +----------------------+ *)

#open "format";;

(*
 * imprime un DL
 *)

let print_dl(dl) =
 
  let premier = ref(true) in (* 1er terme non encore imprim *)

  let print_monme = fun (* imprime un monme *)
  | 0 -> print_string("1")
  | 1 -> print_string("h")
  | e -> print_string("h^"); print_int(e)
  in

  for k=0 to vect_length(dl.coeff)-1 do let c = dl.coeff.(k) and e=k+dl.k0 in

    if c <>. zro then begin (* terme non nul *)

      let s = ref(string_of_nombre c) in

      (* imprime le signe avec des espaces sauf pour le premier terme *)
      if !s.[0] = `-` then begin
         if !premier then print_char `-` else print_string " - ";
         s := sub_string !s 1 (string_length !s - 1)
      end
      else begin
         if !premier then () else print_string " + "
      end;

      (* imprime le coefficient sauf s'il vaut 1 ou -1 *)
      if (!s <> "1") or (e = 0) then print_string(!s);

      (* imprime le monme *)
      if e <> 0 then print_monme(e);
      premier := false;

    end
  done;

  (* imprime le terme complmentaire *)
  if not(!premier) then print_string " + ";
  print_string "o(";
  print_monme(dl.k0 + vect_length(dl.coeff) - 1);
  print_string ")"

;;

#close "format";;

                        (* +-------------------------+
                           |  Impression d'un arbre  |
                           +-------------------------+ *)


#open "arbres";;
#open "format";;

(*
 * convertit une formule en arbre  chaines de caractres
 *)

let rec g_arbre_of_formule(f) = match f with
| Const(c)    -> G_noeud(string_of_nombre(c),[])
| X           -> G_noeud("x",[])
| Fct(nom,g)  -> G_noeud(nom,[g_arbre_of_formule g])
| Bin(op,a,b) -> G_noeud((match op with
                          | Plus  -> "+"
                          | Moins -> "-"
                          | Mult  -> "*"
                          | Div   -> "/"
                          | Puiss -> "^"),
                         [g_arbre_of_formule a; g_arbre_of_formule b])
;;

let print_formule(f) = print_string_g_arbre(g_arbre_of_formule(f));;

#close "format";;


              (* +--------------------------------------------+
                 |  Mise en arbre d'une chane de caractres  |
                 +--------------------------------------------+ *)


#open "genlex";;

let value = 

  (* construit un noeud en valuant les oprations entre constantes *)
  let bin(b, op, x, y) = match x,y with
  | Const(a), Const(b) -> Const(op a b)
  | _                  -> Bin(b, x, y)
  in

  let lexer = make_lexer ["+"; "-"; "*"; "/"; "^"; "x"; "pi"; "e"; "("; ")" ]
  and syntax_of_lexme = function
      | Int n   -> Valeur(Const(nombre_of_int(n)))
      | Float x -> Valeur(Const(nombre_of_float(x)))
      | Kwd "pi"-> Valeur(Const(pi))
      | Kwd "e" -> Valeur(Const(e))
      | Kwd "x" -> Valeur(X)
      | Kwd "+" -> Infixe(1,fun x y -> bin(Plus,  prefix +.,  x,y))
      | Kwd "-" -> Infixe(1,fun x y -> bin(Moins, prefix -.,  x,y))
      | Kwd "*" -> Infixe(2,fun x y -> bin(Mult,  prefix *.,  x,y))
      | Kwd "/" -> Infixe(2,fun x y -> bin(Div,   prefix /.,  x,y))
      | Kwd "^" -> Infixe(3,fun x y -> bin(Puiss, prefix **., x,y))
      | Ident i -> Prfixe(fun x -> Fct(i,x))
      | Kwd "(" -> Lpar
      | Kwd ")" -> Rpar
      | _       -> raise Parse_error
  in let eval = analyseur (fun () -> (2,fun x y -> bin(Mult, prefix *., x,y)))
                          syntax_of_lexme lexer
  in function s -> try eval s
  with Erreur(ex,i) -> print_erreur s i; raise ex
;;

#close "genlex";;
#close "arbres";;

