(*+-------------------------------------------------------------------------+
  |                                                                         |
  |                       Calcul de dveloppements limits                  |
  |                                                                         |
  |                           bibliothque utilitaire                       |
  |                                                                         |
  +-------------------------------------------------------------------------+*)

(* M. Quercia, le 07/12/96 *)

(*
 * Reprsentation des nombres :
 *
 * le type des nombres est dfini de manire abstraite, on dispose des
 * valeurs et fonctions suivantes :
 *)

type nombre;;

value prefix +.   : nombre -> nombre -> nombre  (* addition       *)
  and prefix -.   : nombre -> nombre -> nombre  (* soustraction   *)
  and prefix *.   : nombre -> nombre -> nombre  (* produit        *)
  and prefix /.   : nombre -> nombre -> nombre  (* division       *)
  and prefix **.  : nombre -> nombre -> nombre  (* puissance      *)
  and opp         : nombre -> nombre            (* chg. signe     *)
  and prefix =.   : nombre -> nombre -> bool    (* galit        *)
  and prefix <.   : nombre -> nombre -> bool    (* inf. strict    *)
  and prefix <=.  : nombre -> nombre -> bool    (* inf. large     *)
  and prefix >.   : nombre -> nombre -> bool    (* sup. strict    *)
  and prefix >=.  : nombre -> nombre -> bool    (* sup. large     *)
  and prefix <>.  : nombre -> nombre -> bool    (* non galit    *)
  and is_integer  : nombre -> bool              (* nb entier ?    *)
  and is_rationnal: nombre -> bool              (* nb rationnel ? *)

  and nombre_of_int    : int    -> nombre    (* fonctions de conversion *)
  and nombre_of_float  : float  -> nombre  
  and floor            : nombre -> int       (* partie entire infrieure   *)
  and ceil             : nombre -> int       (* partie entire suprieure   *)
  and string_of_nombre : nombre -> string

  and zro : nombre   (* 0 *)
  and un   : nombre   (* 1 *)
  and deux : nombre   (* 2 *)
  and pi   : nombre   (* 3.14159 *)
  and e    : nombre   (* 2.71828 *)

  and exp : nombre -> nombre  (* fonctions usuelles *)
  and cos : nombre -> nombre
  and sin : nombre -> nombre
  and tan : nombre -> nombre
  and cot : nombre -> nombre
  and asin: nombre -> nombre
  and acos: nombre -> nombre
  and atan: nombre -> nombre
  and ch  : nombre -> nombre
  and sh  : nombre -> nombre
  and th  : nombre -> nombre
  and ash : nombre -> nombre
  and ach : nombre -> nombre
  and ath : nombre -> nombre
  and log : nombre -> nombre
  and sqrt: nombre -> nombre
;;

value calculs_exacts : bool ref;;
(*
 * Si !calculs_exacts vaut true, tous les calculs sur les nombres sont
 * exacts. Un calcul dont le rsultat ne peut pas tre reprsent de manire
 * exacte dclenche l'execption Failure "calcul approch".
 *
 * Si !calculs_exacts vaut false, tout calcul dont le rsultat ne peut tre
 * reprsent de manire exacte produit un nombre approchant ce rsultat.
 * Les seules erreurs pouvant alors survenir sont les calculs impossibles :
 * dpassement de capacit, logarithme d'un nombre ngatif ou nul, ...
 *)

(*
 * Reprsentation d'une formule
 *)

type formule =
| Const of nombre                      (* formule constante *)
| X                                    (* fonction identite *)
| Fct of string * formule              (* fonction compose *)
| Bin of binop * formule * formule     (* opration binaire *)

and binop = Plus | Moins | Mult | Div | Puiss
;;


(* Reprsentation d'un DL : vecteur de coefficients commencant  un indice *)
(* k0 positif ou ngatif. Le coefficient de h^k est :                      *)
(*                                                                         *)
(*    0            si k < k0                                               *)
(*    coeff.(k-k0) si k0 <= k < k0 + vect_length(coeff)                    *)
(*    inconnu      si k >= k0 + vect_length(coeff)                         *)

type dl = { k0: int; coeff: nombre vect};;

(*
 * primitives d'accs aux coefficients :
 *
 * "cre_dl(v,p)" renvoie un dl initialis avec des zros, de valuation "v"
 *               et prcision "p".
 *
 * "valuation(dl)" renvoie le plus petit exposant connu dans "dl".
 * "prcision(dl)" renvoie le plus grand exposant connu dans "dl".
 *
 * "coeff(dl,k)" renvoie le coefficient de h^k dans "dl".
 *               Renvoie 0 si "k < valuation(dl)",
 *               dclenche l'exception Failure "coefficient indtermin"
 *               si "k > prcision(dl)".
 *
 * "store(dl,k,c)" remplace le coefficient de "h^k" dans "dl" par "c".
 *               Dclenche l'exception Failure "exposant invalide" si
 *               "k" n'est pas compris entre "valuation(dl)" et "prcision(dl)".
 *
 * "normalise(dl)" limine les coefficients de tte nuls dans "dl" et renvoie
 *               un dl commenant par un terme non nul ou un dl vide si tous
 *               les coefficients sont nul.
 *
 * "tronque(dl,n)" renvoie la troncature de "dl"  l'ordre "n". Si la prcision
 *               de "dl" est infrieure  "n" dclenche l'exception
 *               Failure "coefficient indtermin"
 *)

value cre_dl   : int * int -> dl;;
value valuation : dl -> int;;
value prcision : dl -> int;;
value coeff     : dl * int -> nombre;;
value store     : dl * int * nombre -> unit;;
value normalise : dl -> dl;;
value tronque   : dl * int -> dl;;

(*
 * Impression formate : les fonctions suivantes sont appeles automatiquement
 * par l'interprteur Caml. Elles ne sont pas conues pour etre utilises
 * directement.
 *)

value print_nombre  : nombre -> unit;;
value print_formule : formule -> unit;;
value print_dl      : dl -> unit;;

(*
 * convertit une chaine de caractres en une formule.
 *)

value value : string -> formule
;;




