(* +------------------------------------------------------------------------+
   |                                                                        |
   |                                                                        |
   |                     Calcul de a^(1/k) en grande prcision              |
   |                                                                        |
   |                                                                        |
   +------------------------------------------------------------------------+ *)

(* traduction caml -> ocaml *)
let make_vect         = Array.create
and sub_vect          = Array.sub
and concat_vect       = Array.append
and vect_length       = Array.length
and float_of_int      = float
and int_of_float      = truncate
and power             = ( ** )
and div_float         = ( /. )
and mult_float        = ( *. )
and concat            = String.concat ""
and printf__sprintf   = Printf.sprintf
and printf__printf    = Printf.printf
and sys__command_line = Sys.argv
;;
(* fin de traduction *)


let base = 10000;;     (* base de numration *)
let mult_limite = 51;; (* en dessous de ce nb de chiffres, mult. ordinaire *)

type grandentier = {
  ch : int array; (* chiffres *)
  lg : int        (* nb de chiffres effectifs *)
};;


                   (* +-------------------------+
                      |  Addition-Soustraction  |
                      +-------------------------+ *)

(* normalise un entier de sorte que le chiffre de tte soit non nul *)
let normalise(e) =
  let i = ref(e.lg) in
  while !i > 0 & e.ch.(!i-1) = 0 do decr i done;
  {ch = e.ch; lg = !i}
;;

(* propage les retenues  la suite d'une addition ou soustraction        *)
(* Dclenche une erreur s'il y a une retenue au del du dernier chiffre. *)
let propage_retenues(e) =
  let c = ref(0) in
  for i=0 to e.lg-1 do
    let x = e.ch.(i) + !c in
    if x >= base  then (e.ch.(i) <- x - base; c := 1)
    else if x < 0 then (e.ch.(i) <- x + base; c := -1)
    else               (e.ch.(i) <- x;        c := 0)
  done;
  if !c <> 0 then failwith "retenue anormale"
;;

(* addition *)
let ( ++ ) e f =
  let n = 1 + max (e.lg) (f.lg) in
  let g = {ch = make_vect n 0; lg = n} in
  for i=0 to e.lg-1 do g.ch.(i) <- e.ch.(i) done;
  for i=0 to f.lg-1 do g.ch.(i) <- g.ch.(i) + f.ch.(i) done;
  propage_retenues(g);
  normalise(g)
;;

(* soustraction *)
let ( -- ) e f =
  let n = 1 + max (e.lg) (f.lg) in
  let g = {ch = make_vect n 0; lg = n} in
  for i=0 to e.lg-1 do g.ch.(i) <- e.ch.(i) done;
  for i=0 to f.lg-1 do g.ch.(i) <- g.ch.(i) - f.ch.(i) done;
  propage_retenues(g);
  normalise(g)
;;

           (* +-----------------------------------------+
              |  Multiplication-Division par un entier  |
              +-----------------------------------------+ *)


let ( *. ) a e =
  let n = e.lg+1 in
  let f = {ch = make_vect n 0; lg = n} in
  let c = ref 0 in
  for i=0 to n-2 do
    let x = a*e.ch.(i) + !c in
    f.ch.(i) <- x mod base;
    c        := x / base
  done;
  f.ch.(n-1) <- !c;
  normalise(f)
;;

let ( /. ) e a =
  let f = {ch = make_vect e.lg 0; lg = e.lg} in
  let c = ref 0 in
  for i=e.lg-1 downto 0 do
    let x = e.ch.(i) + !c*base in
    f.ch.(i) <- x/a;
    c        := x mod a
  done;
  normalise(f)
;;

               (* +---------------------------------+
                  |  Multiplication multiprcision  |
                  +---------------------------------+ *)

(* multiplication ordinaire *)
let mult_n2 e f =
  let n = e.lg + f.lg in
  let g = {ch = make_vect n 0; lg = n} in
  for i=0 to e.lg-1 do
    let c = ref g.ch.(i) in
    for j=0 to f.lg-1 do
      let x = e.ch.(i)*f.ch.(j) + !c in
      g.ch.(i+j) <- x mod base;
      c          := g.ch.(i+j+1) + x/base
    done;
    g.ch.(i+f.lg) <- !c
  done;
  normalise(g)
;;

(* multiplication rcursive : bascule sur la multiplication ordinaire  *)
(* si le nombre de chiffres d'un oprande est infrieur  mult_limite. *)
let rec ( ** ) e f =

  let n = min e.lg f.lg in
  if n <= mult_limite then mult_n2 e f
  else begin
    
    let d = n/2 in
    let (e0,e1) = dcoupe e d
    and (f0,f1) = dcoupe f d in
    let e0f0 = e0 ** f0
    and e1f1 = e1 ** f1 in
    let e2f2 = (e0 ++ e1) ** (f0 ++ f1) -- e0f0 -- e1f1 in

    (* e0f0 < base^(2d) donc e0f0 + base^(2d)e1f1 se rduit  une concatnation *)
    let g = {ch = make_vect (2*d+e1f1.lg) 0; lg=(2*d+e1f1.lg) } in
    for i=0 to e0f0.lg-1 do g.ch.(i)     <- e0f0.ch.(i) done;
    for i=0 to e1f1.lg-1 do g.ch.(2*d+i) <- e1f1.ch.(i) done;

    (* addition " la main" de e2f2*base^d *)
    for i=0 to e2f2.lg-1 do g.ch.(d+i) <- g.ch.(d+i) + e2f2.ch.(i) done;
    propage_retenues(g);
    normalise(g)

  end

and dcoupe e d = (* isole les d premiers chiffres *)
  ({ch = sub_vect e.ch 0 d; lg=d},
   {ch = sub_vect e.ch d (e.lg - d); lg = e.lg - d})
;;

(* multiplication tronque : calcule les d chiffres de poids fort de ef *)
let mult_trq d e f =
  let g = e ** f in
  if g.lg >= d then let (_,res) = dcoupe g (g.lg-d) in res
  else {ch = concat_vect (make_vect (d-g.lg) 0) g.ch; lg = d}
;;

(* calcul de e^k sur d chiffres *)
let rec puiss d e k =
  if k < 1 then failwith "exposant incorrect"
  else if k=1 then e
  else begin
    (* calcule e^(k/2) avec deux chiffres supplmentaires *)
    (* puis lve au carr et multiplie par e avec ces deux chiffres *)
    let f = puiss (d+2) e (k/2) in
    let g = mult_trq (d+2) f f  in
    let h = if k mod 2 = 0 then g else mult_trq (d+2) e g in
    (* retire les deux chiffres en trop *)
    let (_,res) = dcoupe h 2 in res
  end
;;

                          (* +---------------------+
                             |  Calcul de a^(1/k)  |
                             +---------------------+ *)


(* calcule x + x(a-x^k)/(ka). x est suppos avoir 2^d+1 chiffres aprs la *)
(* virgule, le rsultat est donn sur 2^(d+1)+1 chiffres aprs la virgule *)
(* Il est suppos que la partie entire de x^k tient sur un chiffre.      *)
let itre a k x =

  let n = x.lg-1 in (* nb de chiffres aprs la virgule dans x *)
  let p = 2*n-1  in (* nb de chiffres aprs la virgule dans y *)

  (* calcule a - x^k et limine les zros de tte *)
  let xk = puiss (p+3) x k in
  if xk.ch.(xk.lg-1) <> a-1 then failwith "point initial trop loin";
  let i = ref(xk.lg-1) in
  while !i > 0 & xk.ch.(!i-1) = base - 1 do decr i done;
  let a_xk = {ch = make_vect !i 0; lg = !i } in
  for j=0 to !i-1 do a_xk.ch.(j) <- base-1 - xk.ch.(j) done;

  (* multiplie par x/(ka), limine les deux chiffres en trop puis ajoute  x *)
  let (_,z) = dcoupe ((x ** a_xk)/.(k*a)) (n+2) in
  let t = {ch = make_vect (p+1) 0; lg = p+1 } in
  for j=0 to n do t.ch.(p-n+j) <- x.ch.(j) done;
  t ++ z
;;


(* calcule a^(1/k) avec d chiffres aprs la virgule *)
let calcul a k d =

  (* calcul approximatif de a^(1/k), il faut deux chiffres aprs la virgule *)
  let u = power (float_of_int a) (div_float 1.0 (float_of_int k)) in
  let v = mult_float (power (float_of_int base) 2.0) u            in
  let w = ref(int_of_float(v))       in
  let x = ref {ch = [|0;0;0|]; lg=3} in
  for i=0 to 2 do !x.ch.(i) <- !w mod base; w := !w/base done;

  (* itrations *)
  while !x.lg < d+1 do x := itre a k !x done;
  !x
;;


                             (* +--------------+
                                |  Impression  |
                                +--------------+ *)


let string_of_grandentier(e) =
  let s = ref ["\n"] in
  for i=0 to e.lg-1 do
    s := (printf__sprintf " %04d" e.ch.(i)) :: !s;
    if (e.lg-i) mod 15 = 1 then s := "\n" :: !s;
  done;
  concat(!s)  
;;


(* Programme principal *)
begin try
  if vect_length(sys__command_line) <> 4 then failwith "";
  let a = int_of_string(sys__command_line.(1)) in
  let k = int_of_string(sys__command_line.(2)) in
  let d = int_of_string(sys__command_line.(3)) in
  let x = calcul a k d in

  let xk= puiss x.lg x k in
  print_string(string_of_grandentier(x));
  print_newline();
  print_string(string_of_grandentier(xk));
  print_newline();
  ()
with _ ->
  printf__printf "syntaxe : %s a k d\n" sys__command_line.(0);
  flush stdout
end;;
