(* +------------------------------------------------------------------------+
   |                                                                        |
   |                                                                        |
   |                       Calcul de a^(1/k) en grande prcision            |
   |                                                                        |
   |                                                                        |
   +------------------------------------------------------------------------+ *)


open Gmp.F;;
open Infixes;;
let ( -- ) = ui_sub and ( // ) = div_ui;;

let prec = ref(64);; (* nb de bits aprs la virgule *)


(* calcul de x^k *)
let rec puiss k x =
  if k < 1 then failwith "problme"
  else if k=1 then x
  else let y = puiss (k/2) x in if k mod 2 = 1 then y*:y*:x else y*:y
;;

(* calcule x + x(a-x^k)/(ka) *)
let itre a k x =
  prec := 2*(!prec)-32;
  set_precision (!prec+96);
  x +: x*:(a -- (puiss k x))//(k*a)
;;

(* calcule a^(1/k) avec d chiffres aprs la virgule *)
let calcul a k d =

  (* calcul approximatif de a^(1/k), il faut deux chiffres aprs la virgule *)
  let u = (float a) ** (1.0 /. (float k)) in
  let x = ref(from_float u) in

  (* itrations *)
  while !prec < d do x := itre a k !x done;
  !x
;;

(* Programme principal *)
begin try
  if Array.length(Sys.argv) <> 4 then failwith "";
  let a = int_of_string(Sys.argv.(1)) in
  let k = int_of_string(Sys.argv.(2)) in
  let d = int_of_string(Sys.argv.(3)) in
  let d2= truncate(1.0 +. 4.0 *. float(d)/.log10(2.0)) in
  let x = calcul a k d2 in

  let xk= puiss k x in
  print_string(string_from x 10 (4*d+2));
  print_newline();
  print_string(string_from xk 10 (4*d+2));
  print_newline();
  ()
with _ ->
  Printf.printf "syntaxe : %s a k d\n" Sys.argv.(0);
  flush stdout
end;;
