(* +------------------------------------------------------------------------+
   |                                                                        |
   |                                                                        |
   |                     Calcul de a^(1/k) en grande prcision              |
   |                                                                        |
   |                                                                        |
   +------------------------------------------------------------------------+ *)
{ M.Quercia, le 28/04/1998}


const base = 10000;     (* base de numration *)
const mult_limite = 55; (* en dessous de ce nb de chiffres, mult. ordinaire *)

type grandentier = record
		      lg : integer;             (* nb de chiffres effectifs  *)
		      sz : integer;             (* nb de chiffres en mmoire *)
		      ch : array[0..1] of integer; (* chiffres *)
		   end;	 


function max(a,b : integer): integer;
begin
   if a < b then max := b else max := a;
end;

function min(a,b : integer): integer;
begin
   if a < b then min := a else min := b;
end;

                   {+----------------------------+
                    |  Allocation/Dsallocation  |
                    +----------------------------+}


function alloue(sz : integer): ^grandentier;
var p : ^grandentier;
begin
   getmem(p,(sz+2)*sizeof(integer));
   p^.sz := sz;
   alloue := p;
end;

procedure desalloue(p : ^grandentier);
begin
   freemem(p);
end;

                             (* +--------------+
                                |  Impression  |
                                +--------------+ *)


procedure affiche(var e	: grandentier);
var i : integer;
begin
   for i:=e.lg-1 downto 0 do begin
      if      e.ch[i] < 10   then write(' 000',e.ch[i]:0)
      else if e.ch[i] < 100  then write(' 00',e.ch[i]:0)
      else if e.ch[i] < 1000 then write(' 0',e.ch[i]:0)
      else                        write(' ',e.ch[i]:0);
      if ((e.lg-i) mod 15) = 0 then writeln;
   end;
   writeln;
end;

                   (* +-------------------------+
                      |  Addition-Soustraction  |
                      +-------------------------+ *)

(* normalise un entier de sorte que le chiffre de tte soit non nul *)
procedure normalise(var e : grandentier);
var i : integer;
begin
   i := e.lg;
   while (i > 0) and (e.ch[i-1] = 0) do i:=i-1;
   e.lg := i;
end;

(* propage les retenues  la suite d'une addition ou soustraction        *)
(* Dclenche une erreur s'il y a une retenue au del du dernier chiffre. *)
procedure propage_retenues(var e : grandentier);
var c,i,x : integer;
begin
   c := 0;
   for i:=0 to e.lg-1 do begin
      x := e.ch[i] + c;
      if x >= base  then begin e.ch[i] := x - base; c := 1  end
      else if x < 0 then begin e.ch[i] := x + base; c := -1 end
      else               begin e.ch[i] := x;        c := 0  end
  end;
  if c <> 0 then begin writeln("retenue anormale"); halt end;
end;

(* addition : g <- e+f, on peut avoir g == e *)
procedure add(var e,f,g	: grandentier);
var i : integer;
begin
   if @e <> @g then for i:=0 to e.lg-1 do g.ch[i] := e.ch[i];
   for i:=e.lg to f.lg-1  do g.ch[i] := 0;
   for i:=0    to f.lg-1 do g.ch[i] := g.ch[i] + f.ch[i];
   g.lg := 1 + max(e.lg,f.lg);
   g.ch[g.lg-1] := 0;
   propage_retenues(g);
   normalise(g)
end;

(* soustraction : g <- e-f, on peut avoir g == e *)
procedure sub(var e,f,g	: grandentier);
var i : integer;
begin
   if @e <> @g then for i:=0 to e.lg-1 do g.ch[i] := e.ch[i];
   for i:=e.lg to f.lg-1 do g.ch[i] := 0;
   for i:=0    to f.lg-1 do g.ch[i] := g.ch[i] - f.ch[i];
   g.lg := 1 + max(e.lg,f.lg);
   g.ch[g.lg-1] := 0;
   propage_retenues(g);
   normalise(g)
end;

           (* +-----------------------------------------+
              |  Multiplication-Division par un entier  |
              +-----------------------------------------+ *)


(* multiplication : f <- a*e, on peut avoir f == e *)
procedure smul(a : integer; var e,f : grandentier);
var c,i,x : integer;
begin
   c := 0;
   for i:=0 to e.lg-1 do begin
      x := a*e.ch[i] + c;
      f.ch[i] := x mod base;
      c := x div base
   end;
   f.ch[e.lg] := c;
   f.lg := e.lg + 1;
   normalise(f)
end;

(* division : f <- e/a, on peut avoir f == e *)
procedure sdiv(a : integer; var e,f : grandentier);
var c,i,x : integer;
begin
   c := 0;
   for i:=e.lg-1 downto 0 do begin
      x := e.ch[i] + c*base;
      f.ch[i] := x div a;
      c := x mod a
   end;
   f.lg := e.lg;
   normalise(f)
end;

               (* +---------------------------------+
                  |  Multiplication multiprcision  |
                  +---------------------------------+ *)

(* multiplication ordinaire : g <- e*f *)
procedure mult_n2(var e,f,g : grandentier);
var c,i,j,x : integer;
begin
   g.lg := e.lg + f.lg;
   for i:=0 to g.lg-1 do g.ch[i] := 0;
   for i:=0 to e.lg-1 do begin
      c := g.ch[i];
      for j:=0 to f.lg-1 do begin
	 x := e.ch[i]*f.ch[j] + c;
	 g.ch[i+j] := x mod base;
	 c := g.ch[i+j+1] + x/base
      end;
      g.ch[i+f.lg] := c
   end;
   normalise(g)
end;

(* multiplication rcursive : bascule sur la multiplication ordinaire  *)
(* si le nombre de chiffres d'un oprande est infrieur  mult_limite. *)
procedure mult_rec(var e,f,g : grandentier);
var d,i,n		  : integer;
   e0,e1,f0,f1	  : ^grandentier;
   e0f0,e1f1,e2f2 : ^grandentier;
   e0e1,f0f1      : ^grandentier;
begin

   n := min(e.lg,f.lg);
   if n <= mult_limite then mult_n2(e,f,g)
   else begin
    
      d := n div 2;

      e0 := alloue(d); e1 := alloue(e.lg-d);
      for i:=0 to d-1      do e0^.ch[i] := e.ch[i];
      for i:=0 to e.lg-d-1 do e1^.ch[i] := e.ch[d+i];
      e0^.lg := d; e1^.lg := e.lg-d;

      f0 := alloue(d); f1 := alloue(f.lg-d);
      for i:=0 to d-1      do f0^.ch[i] := f.ch[i];
      for i:=0 to f.lg-d-1 do f1^.ch[i] := f.ch[d+i];
      f0^.lg := d; f1^.lg := f.lg-d;

      e0f0 := alloue(e0^.lg+f0^.lg);        mult_n2(e0^,f0^,e0f0^);
      e1f1 := alloue(e1^.lg+f1^.lg);        mult_n2(e1^,f1^,e1f1^);
      e0e1 := alloue(max(e0^.lg,e1^.lg)+1); add(e0^,e1^,e0e1^);
      f0f1 := alloue(max(f0^.lg,f1^.lg)+1); add(f0^,f1^,f0f1^);
      e2f2 := alloue(e0e1^.lg+f0f1^.lg);    mult_n2(e0e1^,f0f1^,e2f2^);
      sub(e2f2^,e0f0^,e2f2^);
      sub(e2f2^,e1f1^,e2f2^);

    (* e0f0 < base^(2d) donc e0f0 + base^(2d)e1f1 se rduit  une concatnation *)
      g.lg := 2*d+e1f1^.lg;
      for i:=0 to e0f0^.lg-1   do g.ch[i] := e0f0^.ch[i];
      for i:=e0f0^.lg to 2*d-1 do g.ch[i] := 0;
      for i:=0 to e1f1^.lg-1   do g.ch[2*d+i] := e1f1^.ch[i];

    (* addition " la main" de e2f2*base^d *)
      for i:=0 to e2f2^.lg-1 do g.ch[d+i] := g.ch[d+i] + e2f2^.ch[i];
      propage_retenues(g);
      normalise(g);

      desalloue(e0);   desalloue(e1);   desalloue(f0);   desalloue(f1);
      desalloue(e0f0); desalloue(e1f1); desalloue(e2f2);
      desalloue(e0e1); desalloue(f0f1);

  end
end;

(* multiplication tronque : calcule les d chiffres de poids fort de e*f *)
procedure mult_trq(d : integer; var e,f,g : grandentier);
var ef : ^grandentier;
   i,j : integer;
begin
   ef := alloue(e.lg+f.lg); mult_rec(e,f,ef^);
   g.lg := d;
   i := ef^.lg-1;
   j := d-1;
   while (i >= 0) and (j >= 0) do begin
      g.ch[j] := ef^.ch[i];
      i := i-1;
      j := j-1;
   end;
   while j >= 0 do begin g.ch[j] := 0; j := j-1 end;
   desalloue(ef);
end;

(* calcul de e^k sur d chiffres *)
procedure puiss(d: integer; var e: grandentier; k: integer; var f: grandentier);
var u,v,w : ^grandentier;
   i	  : integer;
begin
   
  if k < 1 then begin
     writeln("exposant incorrect");
     halt;
  end
  else if k=1 then begin
     f.lg := e.lg;
     for i:=0 to e.lg-1 do f.ch[i] := e.ch[i];
  end
   
  else begin
    (* calcule e^(k/2) avec deux chiffres supplmentaires *)
    (* puis lve au carr et multiplie par e avec ces deux chiffres *)
     u := alloue(d+2); puiss(d+2, e, k div 2, u^);
     v := alloue(d+2); mult_trq(d+2, u^, u^, v^);
     if odd(k) then begin
	mult_trq(d+2,e,v^,u^);
	w := u;
	u := v;
	v := w;
     end;
    (* retire les deux chiffres en trop *)
     f.lg := v^.lg-2;
     for i:=2 to v^.lg-1 do f.ch[i-2] := v^.ch[i];
     desalloue(u); desalloue(v);
  end

end;

                          (* +---------------------+
                             |  Calcul de a^(1/k)  |
                             +---------------------+ *)


(* calcule x + x(a-x^k)/(ka). x est suppos avoir 2^d+1 chiffres aprs la *)
(* virgule, le rsultat est donn sur 2^(d+1)+1 chiffres aprs la virgule *)
(* Il est suppos que la partie entire de x^k tient sur un chiffre.      *)
procedure itere(a,k : integer; var x,y : grandentier);
var i,j,n,p    : integer;
   xk,a_xk,t,z : ^grandentier;
begin

   n := x.lg-1; (* nb de chiffres aprs la virgule dans x *)
   p := 2*n-1;  (* nb de chiffres aprs la virgule dans y *)

  (* calcule a - x^k et limine les zros de tte *)
   xk := alloue(p+3); puiss(p+3, x, k, xk^);
   if xk^.ch[xk^.lg-1] <> a-1 then begin
      writeln("point initial trop loin");
      halt;
   end;
   i := xk^.lg-1;
   while (i > 0) and (xk^.ch[i-1] = base - 1) do i := i - 1;
   a_xk := alloue(i);
   for j:=0 to i-1 do a_xk^.ch[j] := base-1 - xk^.ch[j];
   a_xk^.lg := i;

  (* multiplie par x/(ka), limine les deux chiffres en trop puis ajoute  x *)
   z := alloue(x.lg+a_xk^.lg);
   mult_rec(x,a_xk^,z^);
   sdiv(k*a,z^,z^);
   for j:=n+2 to z^.lg-1 do z^.ch[j-n-2] := z^.ch[j];
   z^.lg := z^.lg-n-2;
   t := alloue(p+1);
   t^.lg := p+1;
   for j:=0 to p-n-1 do t^.ch[j] := 0;
   for j:=0 to n do t^.ch[p-n+j] := x.ch[j];
   add(t^,z^,y);

   desalloue(xk); desalloue(a_xk); desalloue(t); desalloue(z);
end;


(* calcule a^(1/k) avec d chiffres aprs la virgule *)
function calcul(a,k,d :integer) : ^grandentier;
var i,w	: integer;
   x,y	: ^grandentier;
begin
   
  (* calcul approximatif de a^(1/k), il faut deux chiffres aprs la virgule *)
   w := trunc(base*base*exp(ln(a)/k));
   x := alloue(3);
   x^.lg := 3;
   for i:=0 to 2 do begin x^.ch[i] := w mod base; w := w div base end;

  (* itrations *)
   while x^.lg < d+1 do begin
      y := alloue(2*x^.lg-1);
      itere(a,k,x^,y^);
      desalloue(x);
      x := y;
   end;
   calcul := x;
end;



                      {+-----------------------+
                       |  Programme principal  |
                       +-----------------------+}

var a,k,d : integer;
begin
   if paramcount < 3 then begin
      writeln("syntaxe : racine a k d");
      halt;
   end;
   a := atoi(paramstr(1));
   k := atoi(paramstr(2));
   d := atoi(paramstr(3));
   affiche(calcul(a,k,d)^);
end.