(*****************************************************************************
 *                                                                           *
 *                               Jeu du taquin                               *
 *                                                                           *
 * M. Quercia, le 17/10/96                                                   *
 *****************************************************************************)

(* copie de matrices *)
let copy_matrix(mat) =
  let n = vect_length(mat)    in
  let mat' = make_vect n [||] in
  for i=0 to n-1 do mat'.(i) <- copy_vect mat.(i) done;
  mat'
;;


(* affichage d'une matrice *)
#open "format";;
let print_int_matrix(t) =
  let n = vect_length(t) and p = vect_length(t.(0)) in
  for i=0 to n-1 do
    force_newline();
    for j=0 to p-1 do
      let s = ref(string_of_int(t.(i).(j))) in
      while string_length(!s) < 4 do s := " " ^ !s done;
      print_string(!s)
    done
  done;
  force_newline()
;;
#close "format";;
install_printer "print_int_matrix";;

(* animation graphique *)
#open "taquin-tk";;
load_object "taquin-tk";;

(*
 * random_chemin(succ,noeud,n)
 *
 * Effectue n mouvements alatoires sur "noeud" et renvoie le
 * chemin obtenu.
 *)

let random_chemin(succ,noeud,n) =

  (* tire un lment au hasard dans "liste" *)
  let rec random_elt(liste) =
    let k = random__int(list_length liste) in
    list_item k liste
  where rec list_item k liste = match k with
  | 0 -> hd liste
  | _ -> list_item (k-1) (tl liste)
  in

  (* rallonge alatoirement "dbut" de "n" tapes *)
  let rec prolonge(dbut,n) =
    if n=0 then dbut
    else let l = subtract (succ(hd dbut)) dbut in
    (* "l" = liste des successeurs non encore visits *)
    match l with
    | []  -> failwith "impasse"
    | [x] -> prolonge(x::dbut,n-1)
    | _   -> (* s'il y a au moins deux successeurs possibles, tire  *)
             (* au hasard celui  prendre. En cas d'chec, ressaye *)
             (* au plus 2 fois                                      *)
             try prolonge(random_elt(l)::dbut,n-1)
             with Failure "impasse" ->
             try prolonge(random_elt(l)::dbut,n-1)
             with Failure "impasse" ->
             prolonge(random_elt(l)::dbut,n-1)
  
  in prolonge([noeud],n)
;;

