(***********************************************************************)
(*                                                                     *)
(*                           Objective Caml                            *)
(*                                                                     *)
(*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1996 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* $Id: map.ml,v 1.8 1997/10/31 12:59:25 doligez Exp $ *)

(* Modification de M.Quercia le 14/04/98 pour avoir des mappes gnriques *)

type ('a,'b) tree =
    Empty
  | Node of ('a,'b) tree * 'a * 'b * ('a,'b) tree * int
type ('a,'b) t = { cmp:('a->'a->int); tree:('a,'b) tree}

let empty compare = {cmp=compare; tree=Empty}

let height = function
    Empty -> 0
  | Node(_,_,_,_,h) -> h

let create l x d r =
  let hl = height l and hr = height r in
  Node(l, x, d, r, (if hl >= hr then hl + 1 else hr + 1))

let bal l x d r =
  let hl = match l with Empty -> 0 | Node(_,_,_,_,h) -> h in
  let hr = match r with Empty -> 0 | Node(_,_,_,_,h) -> h in
  if hl > hr + 2 then begin
    match l with
      Empty -> invalid_arg "Map.bal"
    | Node(ll, lv, ld, lr, _) ->
        if height ll >= height lr then
          create ll lv ld (create lr x d r)
        else begin
          match lr with
            Empty -> invalid_arg "Map.bal"
          | Node(lrl, lrv, lrd, lrr, _)->
              create (create ll lv ld lrl) lrv lrd (create lrr x d r)
        end
  end else if hr > hl + 2 then begin
    match r with
      Empty -> invalid_arg "Map.bal"
    | Node(rl, rv, rd, rr, _) ->
        if height rr >= height rl then
          create (create l x d rl) rv rd rr
        else begin
          match rl with
            Empty -> invalid_arg "Map.bal"
          | Node(rll, rlv, rld, rlr, _) ->
              create (create l x d rll) rlv rld (create rlr rv rd rr)
        end
  end else
    Node(l, x, d, r, (if hl >= hr then hl + 1 else hr + 1))

let add x data {cmp=compare; tree=tree} =
  let rec add_rec = function
      Empty ->
	Node(Empty, x, data, Empty, 1)
    | Node(l, v, d, r, h) as t ->
	let c = compare x v in
	if c = 0 then
	  Node(l, x, data, r, h)
	else if c < 0 then
	  bal (add_rec l) v d r
	else
	  bal l v d (add_rec r)
  in {cmp=compare; tree=add_rec tree}

let find x  {cmp=compare; tree=tree} =
  let rec find_rec = function
      Empty ->
        raise Not_found
    | Node(l, v, d, r, _) ->
        let c = compare x v in
        if c = 0 then d
        else find_rec (if c < 0 then l else r)
  in find_rec tree

let rec merge t1 t2 =
  match (t1, t2) with
    (Empty, t) -> t
  | (t, Empty) -> t
  | (Node(l1, v1, d1, r1, h1), Node(l2, v2, d2, r2, h2)) ->
      bal l1 v1 d1 (bal (merge r1 l2) v2 d2 r2)

let remove x {cmp=compare; tree=tree} =
  let rec remove_rec = function
      Empty ->
	Empty
    | Node(l, v, d, r, h) as t ->
	let c = compare x v in
	if c = 0 then
	  merge l r
	else if c < 0 then
	  bal (remove_rec l) v d r
	else
	  bal l v d (remove_rec r)
  in {cmp=compare; tree=remove_rec tree}

let iter f {cmp=compare; tree=tree} =
  let rec iter_rec = function
      Empty -> ()
    | Node(l, v, d, r, _) ->
	iter_rec l; f v d; iter_rec r
  in iter_rec tree

let map f {cmp=compare; tree=tree} =
  let rec map_rec = function
      Empty               -> Empty
    | Node(l, v, d, r, h) -> Node(map_rec l, v, f d, map_rec r, h)
  in {cmp=compare; tree=map_rec tree}
 
let fold f {cmp=compare; tree=tree} accu =
  let rec fold_rec m accu =
    match m with
      Empty -> accu
    | Node(l, v, d, r, _) ->
	fold_rec l (f v d (fold_rec r accu))
  in fold_rec tree accu

