{+----------------------------------------------------------------------------+
 |                                                                            |
 |                           Parcours de graphe                               |
 |                                                                            |
 +----------------------------------------------------------------------------+}
{ M.Quercia, le 25/04/98 }

{$i memoire.p}

type pliste = ^cliste;
     cliste = record
		 x    : pointer;
		 suiv : pliste
	      end;

type fdist = function(var x,y): integer;
     fsucc = function(var x): pliste;
     ppoint= procedure(p: pointer);

type graphe = record
		 m    : memoire;  { fonctions de mmorisation            }
		 lg   : integer;  { taille significative d'un noeud      }
		 d    : fdist;    { minore la distance entre deux noeuds }
		 succ : fsucc;    { successeurs d'un noeud               }
		 disp : ppoint;   { dtruit un noeud dj vu             }
	      end;    



                         {+---------------------------+
                          |  Recherche en profondeur  |
                          +---------------------------+}


function pprof(var g: graphe;  var depart,but; pmax: integer): pliste;
var p,q,r : pliste;
   trouve : boolean;
begin

   trouve := egal(depart,but,g.lg);
   p := nil;
   if (not trouve) and (pmax > 0) then begin

      { essaye tous les successeurs }
      q := g.succ(depart);
      while q <> nil do begin

	 if (not trouve) and (g.m.dejavu(g.m.donnees, q^.x^, g.lg) < pmax)
	 then begin
	    { successeur non encore vu ou vu trop loin }
	    g.m.note(g.m.donnees, q^.x^, g.lg, pmax);
	    p := pprof(g, q^.x^,but, pmax-1);
	    trouve := p <> nil;
	    r := q;
	    q := q^.suiv;
	    dispose(r);
	 end

	 else begin
	    { successeur dj vu ou chemin dj trouv }
	    r := q;
	    q := q^.suiv;
	    g.disp(r^.x);
	    dispose(r);
	 end;
	    
      end; { while }

   end;

   if trouve then begin { chemin trouv, ajoute le dpart }
      new(q);
      q^.x := @depart;
      q^.suiv := p;
      pprof := q;
   end
   else pprof := nil;

end; { pprof }
   
	    
                          {+------------------------+
                           |  Recherche en largeur  |
                           +------------------------+}


{ arbre des chemins explors }
type pchemin = ^chemin;
   chemin    = record
		  noeud	: pointer;
		  pere	: pchemin;
		  frere	: pchemin;
	       end;	
   
(* prolonge une liste de chemins d'un cran *)
function avance(var g : graphe; ch: pchemin): pchemin;
var p,s	: pliste;
   q,r	: pchemin;
begin
   r := nil;
   while ch <> nil do begin
      p := g.succ(ch^.noeud^);
      while p <> nil do begin
	 if g.m.dejavu(g.m.donnees, p^.x^, g.lg) < 0
	 then begin
	    g.m.note(g.m.donnees, p^.x^, g.lg, 1);
	    new(q);
	    q^.noeud := p^.x;
	    q^.pere  := ch;
	    q^.frere := r;
	    r := q;
	 end
	 else g.disp(p^.x);
	 s := p;
	 p := p^.suiv;
	 dispose(s);
      end;
      ch := ch^.frere;
   end;
   avance := r;
end; { avance }


(* recherche en largeur *)
function plarg(var g: graphe; var depart,but): pliste;
var p,q	  : pchemin;
   r,s	  : pliste;
   trouve : boolean;
begin

   new(p);
   p^.noeud := @depart;
   p^.pere  := nil;
   p^.frere := nil;
   q := p;
   trouve   := egal(depart,but,g.lg);

   { prolonge les chemins tans qu'on n'est pas au but }
   while not(trouve) do begin
      p := avance(g,p);
      q := p;
      while (q <> nil) and (not trouve) do begin
	 trouve := egal(q^.noeud^,but,g.lg);
	 if not trouve then q := q^.frere;
      end;
   end;

   { but atteint en q^, reconstitue le chemin }
   r := nil;
   while q <> nil do begin
      new(s);
      s^.x := q^.noeud;
      s^.suiv := r;
      r := s;
      q := q^.pere;
   end;

   plarg := r;

end; { plarg }


(* largeur par les deux bouts *)
function plarg2(var g1,g2 : graphe; var depart,but): pliste;
var p1,p2,q1,q2	: pchemin;
   r,s,t	: pliste;
   trouve	: boolean;
   phase	: boolean;
begin

   new(p1);
   p1^.noeud := @depart;
   p1^.pere  := nil;
   p1^.frere := nil;
   q1 := p1;

   new(p2);
   p2^.noeud := @but;
   p2^.pere  := nil;
   p2^.frere := nil;
   q2 := p2;

   trouve := egal(depart,but,g1.lg);
   phase  := true;

   { avance alternativement les chemins de p1 et p2 }
   while not trouve do begin

      if phase then p1 := avance(g1,p1) else p2 := avance(g2,p2);
      phase := not phase;
      
      q1 := p1;
      while (q1 <> nil) and (not trouve) do begin
	 q2 := p2;
	 while (q2 <> nil) and (not trouve) do begin
	    trouve := egal(q1^.noeud^,q2^.noeud^,g1.lg);
	    if not trouve then q2 := q2^.frere;
	 end;
	 if not trouve then q1 := q1^.frere;
      end;

   end;

   { point de rencontre dans q1^ et q2^, reconstitue le chemin }
   new(r);
   t := r;
   r^.x := q2^.noeud;
   q2 := q2^.pere;
   while q2 <> nil do begin
      new(r^.suiv);
      r := r^.suiv;
      r^.x := q2^.noeud;
      q2 := q2^.pere;
   end;
   r^.suiv := nil;
   q1 := q1^.pere;
   while q1 <> nil do begin
      new(s);
      s^.x := q1^.noeud;
      s^.suiv := t;
      t := s;
      q1 := q1^.pere;
   end;
   plarg2 := t;

end; { plarg2 }
   
   
                   {+--------------------------------------+
                    |  Recherche heuristiquement ordonne  |
                    +--------------------------------------+}



{ chemins classs par cot croissant }
type pvchemin = ^vchemin;
   vchemin    = record
		   v  : integer;    { longueur estime  }
		   l  : integer;    { longueur atteinte }
		   ch : pliste;     { chemin invers    }
		   nx : pvchemin;   { -> chemin suivant }
		end;  

function pheur(var g: graphe; var depart,but): pliste;
var p,q,z : pvchemin;
   r,s,t  : pliste;
   u	  : ^pvchemin;
   l,v	  : integer;
begin

   new(p);
   p^.v := g.d(depart,but);
   p^.l := 0;
   new(p^.ch);
   p^.ch^.x := @depart;
   p^.ch^.suiv  := nil;
   p^.nx := nil;

   while not egal(p^.ch^.x^,but,g.lg) do begin

      { dveloppe le meilleur chemin }
      r := p^.ch;
      l := p^.l + 1;
      q := p;
      p := p^.nx;
      dispose(q);

      { insre chaque successeur dans la liste trie }
      s := g.succ(r^.x^);
      while s <> nil do begin

	 t := s^.suiv;
	 s^.suiv := r; { lie le noeud  son pre }
	 v := g.m.dejavu(g.m.donnees,s^.x^,g.lg);

	 if v > l then begin { noeud dj vu plus loin, supprime le mauvais }
	    u := @p;
	    while true do
	       if egal(u^^.ch^.x^,s^.x^,g.lg) then begin
		  z := u^;
		  u^ := z^.nx;
		  dispose(z);
		  leave;
	    { Rmq: il n'est pas prudent de dsallouer le chemin z^.ch }
	    { car il peut y avoir des chemins qui pointent dessus.    }
	    { Une meilleure solution serait d'utiliser une table      }
	    { d'association noeud -> meilleur pre, mail il faudrait  }
	    { rcrire le module mmoire ...                          }
	       end
	       else u := @(u^^.nx);
	 end;

	 if (v < 0) or (v > l) then begin { noeud non encore vu ou amlioration }
	    g.m.note(g.m.donnees,s^.x^,g.lg,l);
	    v := l + g.d(s^.x^,but);
	    u := @p;
	    while (u^ <> nil) and (u^^.v < v) do u := @(u^^.nx);
	    new(q);
	    q^.v  := v;
	    q^.l  := l;
	    q^.ch := s;
	    q^.nx := u^;
	    u^ := q;
	 end
	 else begin { noeud sans intrt }
	    g.disp(s^.x);
	    dispose(s);
	 end;

	 s := t;
      end; { while s <> nil }

   end; { while not egal }

   { retourne le chemin trouv }
   r := nil;
   s := p^.ch;
   while s <> nil do begin
      t := s^.suiv;
      s^.suiv := r;
      r := s;
      s := t;
   end;
   pheur := r;

end; { pheur }


                         {+---------------------------+
                          |  Tire un noeud au hasard  |
                          +---------------------------+}

function random_noeud(var g: graphe; var depart; pmax: integer): pointer;
var p,q,r : pliste;
   s	  : pointer;
   i,n	  : integer;
begin

   if pmax <= 0 then random_noeud := @depart
   else begin

      { cherche les successeurs non encore vus }
      p := g.succ(depart);
      n := 0;
      q := nil;
      while p <> nil do begin
	 r := p;
	 p := p^.suiv;
	 if g.m.dejavu(g.m.donnees,r^.x^,g.lg) > 0 then begin
	    g.disp(r^.x);
	    dispose(r);
	 end
	 else begin
	    r^.suiv := q;
	    q := r;
	    n := n+1;
	 end;
      end;
      if n = 0 then random_noeud := nil

      { en esaye un au hasard }
      else begin
	 p := q;
	 for i:=1 to random(n) do p := p^.suiv;
	 s := random_noeud(g, p^.x^, pmax-1);
	 if s = nil then begin { loup, rssaye }
	    p := q;
	    for i:=1 to random(n) do p := p^.suiv;
	    s := random_noeud(g, p^.x^, pmax-1);
	 end;
	 random_noeud := s;
      end;
   end;
end; { random_noeud }
   
	    
      
	    