{+----------------------------------------------------------------------------+
 |                                                                            |
 |                        Fonctions de Mmorisation                           |
 |                                                                            |
 +----------------------------------------------------------------------------+}
{ M.Quercia, le 25/04/98 }


type pointer = ^byte;
     abyte   = packed array[0..1] of byte;
     pabyte  = ^abyte;
     fvint   = function(p: pointer; var x; lg: integer): integer;
     pvint   = procedure(var p: pointer; var x; lg: integer; n: integer);
     fint    = function(p: pointer): integer;
     pvoid   = procedure(var p: pointer);
   
type memoire = record
		  donnees : pointer;
		  dejavu  : fvint;
		  note	  : pvint;
		  taille  : fint;
		  raz	  : pvoid;
		 end;	  

{
 mem.donnees = pointeur vers la structure de donnes prive de la mmoire.
 mem.dejavu(p,x,lg) retourne le nombre mmoris pour "x" ou -1 si "x" n'est
    pas en mmoire. "p" = pointeur donnes, "lg" = longueur de "x" en octets.
 mem.note(p,x,lg,n) enregistre "x" avec le nombre "n" (doit tre positif).
 mem.taille(p) retourne le nombre d'lments en mmoire.
 mem.raz(p) libre la mmoire.
 }
		  

                         {+--------------------------+
                          |  Comparaison en mmoire  |
                          +--------------------------+}


function egal(var x,y; lg:integer): boolean;
var i : integer;
begin
   i := 0;
   while (i < lg) and (abyte(x)[i] = abyte(y)[i]) do i := i+1;
   egal := i = lg;
end; { egal }
   

type pcellule = ^cellule;
   cellule    =  record
		    x	 : pabyte;        { lment mmoris }
		    n	 : integer;       { nombre associ   }
		    frere: pcellule;      { cellule suivante }
		    fils : pcellule;      { pour arbre       }
		 end;	 
   

                     {+----------------------------------+
                      |  Mmorisation par liste chaine  |
                      +----------------------------------+}

function l_dejavu(p: pointer; var x; lg: integer): integer;
var n : integer;
   q  : pcellule;
begin
   q := pcellule(p);
   n := -1;
   while (q <> nil) and (n < 0) do begin
      if egal(q^.x^, x, lg) then n := q^.n;
      q := q^.frere;
   end;
   l_dejavu := n;
end; { l_dejavu }


procedure l_note(var p: pointer; var x; lg: integer; n: integer);
var q : pcellule;
begin
   new(q);
   q^.x := @x;
   q^.n := n;
   q^.frere := p;
   p := q;
end;

function l_taille(p: pointer): integer;
var n : integer;
   q  : pcellule;
begin
   q := pcellule(p);
   n := 0;
   while q <>nil do begin
      n := n+1;
      q := q^.frere;
   end;
   l_taille := n;
end; { l_taille }

procedure l_raz(var p: pointer);
var q : pcellule;
   r  : pcellule;
begin
   q := pcellule(p);
   while q <> nil do begin
      r := q;
      q := q^.frere;
      dispose(r);
   end;
   p := nil;
end; { l_raz }
   
procedure memoire_a_liste(var m: memoire);
begin with m do begin
   donnees := nil;
   dejavu  := l_dejavu;
   note    := l_note;
   taille  := l_taille;
   raz     := l_raz;
end end;



                     {+----------------------------------+
                      |  Mmorisation par arbre digital  |
                      +----------------------------------+}


function a_dejavu(p : pointer; var x; lg: integer): integer;
var n,i	: integer;
   q	: pcellule;
begin

   q := pcellule(p);
   i := 0;
   n := -1;
   
   while q <> nil do begin
      if q^.x^[i] = abyte(x)[i] then begin
	 if (q^.fils = nil) and egal(q^.x^,x,lg) then n := q^.n;
	 i := i+1;
	 q := q^.fils;
      end
      else q := q^.frere;
   end;

   a_dejavu := n;
end; { a_dejavu }
   
procedure a_note(var p : pointer; var x; lg: integer; n: integer);
var i : integer;
   q  : pcellule;
begin
		    
   q := pcellule(p);
   if q = nil then begin
      new(q);
      q^.x := @x;
      q^.n := n;
      q^.fils := nil;
      q^.frere := nil;
      p := q
   end
   
   else begin
      i := 0;
      while q <> nil do begin
	 if q^.x^[i] = abyte(x)[i] then begin
	    if q^.fils = nil
	       then if egal(q^.x^,x,lg)
		  then q^.n := n
	       else begin
		  new(q^.fils);
		  q^.fils^.x := q^.x;
		  q^.fils^.n := q^.n;
		  q^.fils^.fils  := nil;
		  q^.fils^.frere := nil;
	       end;
	    i := i+1;
	    q := q^.fils;
	 end
	 else begin
	    if q^.frere = nil then begin
	       new(q^.frere);
	       q^.frere^.x := @x;
	       q^.frere^.n := n;
	       q^.frere^.fils  := nil;
	       q^.frere^.frere := nil;
	       q := nil;
	    end
	    else q := q^.frere;
	 end;
      end;
   end;
end; { a_note }
   

function a_taille(p : pointer): integer;
var q : pcellule;
begin
   q := pcellule(p);
   if q = nil then a_taille := 0
   else if q^.fils = nil then a_taille := 1 + a_taille(q^.frere)
   else a_taille := a_taille(q^.fils) + a_taille(q^.frere);
end; { a_taille }
   
procedure a_raz(var p : pointer);
var q : pcellule;
begin
   q := pcellule(p);
   if q <> nil then begin
      a_raz(pointer(q^.fils));
      a_raz(pointer(q^.frere));
      dispose(q);
   end;
   p := nil;
end; { a_raz }

procedure memoire_a_arbre(var m	: memoire);
begin with m do begin
   donnees := nil;
   dejavu  := a_dejavu;
   note    := a_note;
   taille  := a_taille;
   raz     := a_raz;
end end;

