(*--------------------------------------------------------------------------+
 |                                                                          |
 |               Bote  outils pour les formules booleennes                |
 |                                                                          |
 |  M. Quercia, le 08/09/96                                                 |
 +--------------------------------------------------------------------------+*)

#open "arbres";;
load_object "arbres";;

type expression =
| Const of bool
| Var of string
| Neg of expression
| Bin of bin_op * expression * expression

and bin_op = Et | Ou | Oubien | Equiv | Impl
;;


                         (* +-----------------------+
                            |  Oprateurs associs  |
                            +-----------------------+ *)

let opr = fun
| Et     -> prefix &
| Ou     -> prefix or
| Oubien -> prefix <>
| Equiv  -> prefix =
| Impl   -> prefix <=
;;

              (* +--------------------------------------------+
                 |  Mise en arbre d'une chane de caractres  |
                 +--------------------------------------------+ *)


#open "genlex";;

let expression =
  let lexer = make_lexer
    ["vrai"; "faux"; "non"; "et"; "ou"; "oubien"; "<=>"; "=>"; "<="; "("; ")"]
  and syntax_of_lexme = function
  | Kwd "vrai"   -> Valeur(Const true)
  | Kwd "faux"   -> Valeur(Const false)
  | Kwd "non"    -> Prfixe(fun x -> Neg(x))
  | Kwd "et"     -> Infixe(3,fun x y -> Bin(Et,x,y))
  | Kwd "ou"     -> Infixe(2,fun x y -> Bin(Ou,x,y))
  | Kwd "oubien" -> Infixe(2,fun x y -> Bin(Oubien,x,y))
  | Kwd "<=>"    -> Infixe(1,fun x y -> Bin(Equiv,x,y))
  | Kwd "=>"     -> Infixe(1,fun x y -> Bin(Impl,x,y))
  | Kwd "<="     -> Infixe(1,fun x y -> Bin(Impl,y,x))
  | Kwd "("      -> Lpar
  | Kwd ")"      -> Rpar
  | Ident(x)     -> Valeur(Var(x))
  | _            -> raise Parse_error
  in let eval = analyseur (fun () -> raise Parse_error) syntax_of_lexme lexer
  in fun s -> try eval s
              with Erreur(exn,i) -> print_erreur s i; raise exn
;;

#close "genlex";;



                              (* +-------------+
                                 |  Affichage  |
                                 +-------------+ *)


let rec garb e = match e with
| Const(true)  	  -> G_noeud("vrai",[])
| Const(false) 	  -> G_noeud("faux",[])
| Var(v)       	  -> G_noeud(v,[])
| Neg(u)       	  -> G_noeud("non",[garb u])
| Bin(Et,u,v)     -> G_noeud("et",[garb u; garb v])
| Bin(Ou,u,v)     -> G_noeud("ou",[garb u; garb v])
| Bin(Oubien,u,v) -> G_noeud("oubien",[garb u; garb v])
| Bin(Equiv,u,v)  -> G_noeud("<=>",[garb u; garb v])
| Bin(Impl,u,v)   -> G_noeud("=>",[garb u; garb v])
;;

let print_exp e = print_string_g_arbre(garb e);;
install_printer "print_exp";;




                          (* +---------------------+
                             |  Formule alatoire  |
                             +---------------------+ *)

(* random_exp h n tire une formule au hasard comportant au plus n variables
   et de hauteur <= h. Les noms des variables sont a,b,c,... donc il est
   souhaitable mais non impos que n <= 26. *)
let rec random_exp =
  let random_var(n) = Var(make_string 1 (char_of_int(97+random__int(n))))
  and bin(r) = match r with 0 -> Et | 1 -> Ou | 2 -> Oubien | 3 -> Equiv | _ -> Impl
  in fun h n ->
  if h <= 0 then random_var(n)
  else match random__int(7) with
  | 0 -> random_var(n)
  | 1 -> Neg(random_exp (h-1) n)
  | r -> Bin(bin(r-2),random_exp (h-1) n, random_exp (h-1) n)
;;

#close "arbres";;
