{+----------------------------------------------------------------------------+
 |                                                                            |
 |                arbres.p : affichage semi-graphique d'arbres                |
 |                                                                            |
 +----------------------------------------------------------------------------+}
{ M.Quercia, le 12/04/98 }

{ Utilisation :

  Dfinir une fonction produisant un arbre  tiquettes chaines de caractres
  (voir le type ci-dessous)  partir d'un arbre abstrait. Seuls les champs
  "etiq", "fils", "frere" doivent renseigns.

  La procdure "imprime_arbre" calcule les positions d'affichage de chaque
  noeud, puis affiche l'arbre.

  La procdure "detruit_arbre" dsalloue tous les noeuds d'un arbre.

}
 
type
   parbre = ^arbre;
   arbre  = record
	       etiq  : string;   { lgende                   }
	       pos   : integer;  { abscisse relative au pre }
	       tg    : integer;  { trait gauche              }
	       td    : integer;  { trait droit               }
	       fils  : parbre;   { premier fils              }
	       frere : parbre;   { frres suivants           }
               suiv  : parbre;   { successeur en largeur     }
	    end;     

      {+-------------------------------------------------------+
       |  Remplit rcursivement les champs "pos" "tg" et "td"  |
       +-------------------------------------------------------+}

{ x0 = abscisse mini utilisable (rw) }
{ x2 <- abscisse dernier frre       }
procedure calcule_pos(var a:arbre; var x0,x2:integer);
var lg,lf,d,e,l,x1 : integer;
   p		   : parbre;
begin

   { traitement des fils }
   l := length(a.etiq);
   if a.fils <> nil then begin

      { calcule les positions relatives }
      lg := 0;
      calcule_pos(a.fils^,lg,x2);
      x1 := a.fils^.pos;

      { centre les fils par rapport au pre }
      d := (x1 + x2) div 2;
      p := a.fils;
      while p <> nil do begin
	 p^.pos := p^.pos - d;
	 p := p^.frere;
      end;
      
      { centre l'tiquette }
      lf := x2-x1-3;
      e  := lf-l;
      if e >= 0 then begin
	 a.tg  := e div 2;
	 a.td  := e-a.tg;
      end
      else begin { tiquette trop grande }
	 a.tg  := 0;
	 a.td  := 0;
	 lf    := l;
      end;
      if d > (lf div 2) then a.pos := x0 + d else a.pos := x0 + (lf div 2);
      if lf > lg then x0 := x0 + lf else x0 := x0 + lg;

   end
   else begin { pas de fils }
      a.tg  := 0;
      a.td  := 0;
      a.pos := x0 + (l div 2);
      x0    := x0+l;
   end;
      

   { traitement des frres }
   if a.frere <> nil then begin
      x0 := x0+2;
      calcule_pos(a.frere^,x0,x2);
   end
   else x2 := a.pos;

end; { calcule_pos }



                            {+--------------------+
                             |  Imprime un arbre  |
                             +--------------------+}

procedure imprime_arbre(var a : arbre);
var i,l,x,y,z : integer;
   p,r,t      : parbre;
   q,s	      : ^parbre;
begin

   x := 0;
   calcule_pos(a,x,y);
   a.suiv := nil;
   
   new(q);
   q^ := @a;
   while q^ <> nil do begin

      { imprime les tiquettes et les traits }
      x := 0;
      p := q^;
      while p <> nil do with p^ do begin
	 l := length(etiq) + tg + td;
	 for i := x to pos - ((l-1) div 2) - 1 do write(' ');
	 for i := 1 to tg do write('_');
	 write(etiq);
	 for i := 1 to td do write('_');
	 x := pos + (l div 2)  + 1;
	 p := suiv;
      end;
      writeln;

      { imprime les bouts de branche et concatne les fils }
      x := 0;
      p := q^;
      s := @t;
      while p <> nil do begin
	 s^ := p^.fils;
	 while s^ <> nil do with s^^ do begin
	    y := pos + p^.pos;
	    if pos < 0 then z := y
	    else if pos = 0 then z := y - 1
	    else z := y - 2;
	    for i := x to z do write(' ');
	    if pos < 0 then write('/')
	    else if pos = 0 then write('|')
	    else write('\');
	    x := z+2;
	    pos := y;
	    suiv := frere;
	    s := @suiv;
	 end;
	 r := p;
	 p := p^.suiv;
      end;
      s^ := nil;
      writeln;

      { et recommence avec les fils }
      q^ := t;

   end;

end; { imprime_arbre }


                            {+--------------------+
                             |  Dtruit un arbre  |
                             +--------------------+}

procedure detruit_arbre(p : parbre) ;
begin if p <> nil then with p^ do begin
   detruit_arbre(fils);
   detruit_arbre(frere);
   dispose(p);
end end;

