{+----------------------------------------------------------------------------+
 |                                                                            |
 |                          parser.p : analyseur syntaxique                   |
 |                                                                            |
 +----------------------------------------------------------------------------+}
{ M.Quercia le 12/04/98 }

{ Utilisation :

 Dfinir le type "value" avant d'inclure ce fichier.
 On dispose alors de la fonction "analyse" qui prend en entre une
 chaine de caractres et un "lexeur" et retourne la valeur reprsente
 par cette chaine en tenant compte des priorits et des parenthses.

 "lexeur" est une fonction appele pour chaque "mot" isol dans
 la chaine d'entre et qui doit retourner le lexme correspondant 
 ce mot (voir la dfinition du type "lexeme" ci-dessous).
 Le paramtre "infx" indique si on attend un oprateur infixe ou non,
 il peut tre utilis pour diffrencier le moins unaire et le moins
 binaire par exemple.

 Un mot est soit :
   une parenthse (non soumis au lexeur)
   une suite de caractres opratoires (+,-,*,/,=,<,>,..)
   une suite de chiffres
   une suite de lettres ou chiffres commenant par une lettre.

 Les lexmes de type Infixe, Prefixe, Postfixe contiennent une fonction
  appeler lorsque l'on dispose du ou des oprandes (champs pr, po, ix).
 Les lexmes Infixe contiennent aussi un niveau de priorit (champ p);
 les oprations infixes non parenthses sont effectues par ordre de
 priorits dcroissantes et de gauche  droite en cas de priorits gales.
 
 En cas d'erreur de syntaxe, "analyse" affiche un message d'erreur
 et arrte le programme.

}

type f_value  = function(x : value)   : value;
     ff_value = function(x,y : value) : value;
     
type t_lexeme = (Valeur, Prefixe, Postfixe, Infixe, Lpar, Rpar);
     lexeme = record case nature: t_lexeme of
		Valeur	 : (va : value);
		Prefixe	 : (pr : f_value);
		Postfixe : (po : f_value);
		Infixe	 : (ix : ff_value; p : integer);
              end; { case }

type t_lexer = procedure(s : string; infx : boolean; var lx : lexeme);


                    {+---------------------------+
                     |  Dlimite le mot suivant  |
                     +---------------------------+}

const pmax = 50; { niveau maximum d'imbrication }
   
function analyse(s : string; lexeur : t_lexer) : value;

  var i,j : integer;         { indices mot suivant dans s }
     l	  : integer;                  { longueur de s     }
     lx	  : lexeme;                   { lexme courant    }
     pile : array[0..pmax] of lexeme; { pile d'valuation }
     p	  : integer;                  { pointeur de pile  }

  { classification sommaire des caractres }
  function categorie(c : char) : integer;
  begin
     if c < ' '       then categorie := 0  { invalide   }
     else if c = ' '  then categorie := 1  { espace     }
     else if c = '('  then categorie := 2  { parenthse }
     else if c = ')'  then categorie := 2  { parenthse }
     else if c < '0'  then categorie := 3  { oprateur  }
     else if c <= '9' then categorie := 4  { chiffre    }
     else if c < 'A'  then categorie := 3  { oprateur  }
     else                  categorie := 5; { lettre     }
  end; { categorie }

  { dlimite le mot suivant }
  procedure mot_suivant;
  begin

     i := j;
     while (i <= l) and (categorie(s[i]) <= 1) do i := i+1;
     j := i+1;

     if i <= l then case categorie(s[i]) of
       2 : ;
       3 : while (j <= l) and (categorie(s[j]) =  3) do j := j+1;
       4 : while (j <= l) and (categorie(s[j]) =  4) do j := j+1;
       5 : while (j <= l) and (categorie(s[j]) >= 4) do j := j+1;
     end; { case }

  end;

  { erreur de syntaxe }
  procedure erreur;
  begin
     writeln('erreur de syntaxe caractre ', i);
     halt;
  end; { erreur }

  { empile un lexme }
  procedure empile;
  begin case lx.nature of
    
    Valeur   : if (p >= 0) and (pile[p].nature = Prefixe)
	       then begin lx.va := pile[p].pr(lx.va); p := p-1; empile end
	       else begin p := p+1; pile[p] := lx end;

    Prefixe  : begin p := p+1; pile[p] := lx end;
    

    Postfixe : if (p >= 0) and (pile[p].nature = Valeur)
	       then pile[p].va := lx.po(pile[p].va)
	       else erreur;

    Infixe   : if (p >= 0) and (pile[p].nature = Valeur)
	       then  if (p >= 1)
		    and (pile[p-1].nature = Infixe)
		    and (pile[p-1].p >= lx.p) 
		    then begin
		       pile[p-2].va := pile[p-1].ix(pile[p-2].va,pile[p].va);
		       p := p-2;
		       empile;
		     end
		     else begin p := p+1; pile[p] := lx end
	       else erreur;

    Lpar     : begin p := p+1; pile[p] := lx end;

    Rpar     : if (p >= 1) and (pile[p].nature = Valeur)
	       then case pile[p-1].nature of
                 Infixe	: begin
		            pile[p-2].va := pile[p-1].ix(pile[p-2].va,pile[p].va);
		            p := p-2;
		            empile;
			  end;
		 Lpar	: begin
			     lx := pile[p];
			     p := p-2;
			     empile;
			  end;
  	         else erreur;
	       end
	       else erreur;

  end end;

begin { fonction analyse }

   p := 0; pile[0].nature := Lpar;
   j := 0; l := length(s);
   mot_suivant;

   while (i <= l) and (p >= 0) do begin
      case s[i] of
	'(' : lx.nature := Lpar;
	')' : lx.nature := Rpar;
        else lexeur(copy(s,i,j-i), pile[p].nature = Valeur, lx);
      end;
      empile;
      mot_suivant;
   end;
   lx.nature := Rpar; empile;
   if p <> 0 then erreur;
   analyse := pile[0].va;

end; { analyse }
  
