{+----------------------------------------------------------------------------+
 |                                                                            |
 | tp-bool.p: Fonctions et procdures utilitaires pour le TP tautologies      |
 |                                                                            |
 +----------------------------------------------------------------------------+}
{ M.Quercia, le 12/04/98 }

{ Objets exports :

  type "formule"
  procedure "imprime_formule" : affiche une formule
  fonction "evalue" : valuation d'une chaine de caractres
  fonction "f_alea" : retourne une formule alatoire
  fonction "copie_formule" : duplique une formule
 
}
 

type
   bool_op   = (Et, Ou, Oubien, Equiv, Impl);
   bool_node = (Constante, Variable, Non, Bin);
   pformule  = ^formule;
   formule   = record case nature : bool_node of
		 Constante : (c : boolean);
		 Variable  : (v : string);
		 Non	   : (n : pformule);
		 Bin	   : (op : bool_op;
			      g  : pformule;
			      d  : pformule);
	       end; { case }				 

           {+--------------------------------------------+
            |  Mise en arbre d une chaine de caractres  |
            +--------------------------------------------+}

type value = pformule;		  
{$i parser.p}   { analyseur syntaxique }

function formule_et(x,y	: pformule): pformule;
var p : pformule;
begin
   new(p);
   p^.nature := Bin;
   p^.op := Et;
   p^.g := x;
   p^.d := y;
   formule_et := p;
end; { formule_et }
   
function formule_ou(x,y	: pformule): pformule;
var p : pformule;
begin
   new(p);
   p^.nature := Bin;
   p^.op := Ou;
   p^.g := x;
   p^.d := y;
   formule_ou := p;
end; { formule_ou }
   
function formule_oubien(x,y : pformule): pformule;
var p : pformule;
begin
   new(p);
   p^.nature := Bin;
   p^.op := Oubien;
   p^.g := x;
   p^.d := y;
   formule_oubien := p;
end; { formule_oubien }
   
function formule_equiv(x,y : pformule): pformule;
var p : pformule;
begin
   new(p);
   p^.nature := Bin;
   p^.op := Equiv;
   p^.g := x;
   p^.d := y;
   formule_equiv := p;
end; { formule_equiv }
   
function formule_impl(x,y : pformule): pformule;
var p : pformule;
begin
   new(p);
   p^.nature := Bin;
   p^.op := Impl;
   p^.g := x;
   p^.d := y;
   formule_impl := p;
end; { formule_impl }

function formule_non(x : pformule): pformule;
var p : pformule;
begin
   new(p);
   p^.nature := Non;
   p^.n := x;
   formule_non := p;
end; { formule_non }

procedure lexeur(s : string; infx : boolean; var lx : lexeme);
begin

   if s = 'vrai' then begin
      lx.nature := Valeur;
      new(lx.va);
      lx.va^.nature := Constante;
      lx.va^.c := true;
   end

   else if s = 'faux' then begin
      lx.nature := Valeur;
      new(lx.va);
      lx.va^.nature := Constante;
      lx.va^.c := false;
   end

   else if s = 'non' then begin
      lx.nature := Prefixe;
      lx.pr := formule_non;
   end

   else if s = 'et' then begin
      lx.nature := Infixe;
      lx.ix := formule_et;
      lx.p  := 2;
   end
   
   else if s = 'ou' then begin
      lx.nature := Infixe;
      lx.ix := formule_ou;
      lx.p  := 1;
   end

   else if s = 'oubien' then begin
      lx.nature := Infixe;
      lx.ix := formule_oubien;
      lx.p  := 1;
   end

   else if s = '<=>' then begin
      lx.nature := Infixe;
      lx.ix := formule_equiv;
      lx.p  := 0;
   end

   else if s = '=>' then begin
      lx.nature := Infixe;
      lx.ix := formule_impl;
      lx.p  := 0;
   end

   else begin
      lx.nature := Valeur;
      new(lx.va);
      lx.va^.nature := Variable;
      lx.va^.v := s;
   end;

end; { lexeur }

function evalue(s : string) : pformule;
begin
   evalue := analyse(s,lexeur);
end; { evalue }


  {+--------------------------------------------------------------+
   |  Construit une formule alatoire  n variables de hauteur h  |
   +--------------------------------------------------------------+}

function f_alea(n,h : integer) : pformule;
var r : integer;
   p  : pformule;
begin

   new(p);
   if h <= 0 then begin
      p^.nature := Variable;
      p^.v := char(97+random(n));
   end
   else begin
      r := random(7);
      case r of
	0 : begin
	       p^.nature := Variable;
	       p^.v := char(97+random(n));
	    end;
	1 : begin
	       p^.nature := Non;
	       p^.n := f_alea(n,h-1);
	    end;
      else begin
	       p^.nature := Bin;
	       case r of
	 	 2 : p^.op := Et;
	 	 3 : p^.op := Ou;
	 	 4 : p^.op := Oubien;
	 	 5 : p^.op := Equiv;
	 	 6 : p^.op := Impl;
	       end; { case }
	       p^.g := f_alea(n,h-1);
	       p^.d := f_alea(n,h-1);
            end;
      end; { case }
   end;
   
   f_alea := p;
   
end;


                      {+-----------------------+
                       |  Affiche une formule  |
                       +-----------------------+}

{$i arbres.p}  { affichage d'arbres }

function arbre_of_formule(f : formule) : parbre;
var p : parbre;
begin
   new(p);
   case f.nature of
     Constante : begin
		    if f.c then p^.etiq := "vrai" else p^.etiq := "faux";
		    p^.fils  := nil;
		 end;
     Variable  : begin
		    p^.etiq  := f.v;
		    p^.fils  := nil;
		 end;
     Non       : begin
		    p^.etiq := "non";
		    p^.fils := arbre_of_formule(f.n^);
		 end;
     Bin       : begin
		    case f.op of
		      Et     : p^.etiq := "et";
		      Ou     : p^.etiq := "ou";
		      Oubien : p^.etiq := "oubien";
		      Equiv  : p^.etiq := "<=>";
		      Impl   : p^.etiq := "=>";
		    end; { case }
		    p^.fils := arbre_of_formule(f.g^);
		    p^.fils^.frere := arbre_of_formule(f.d^);
		 end;
   end; { case }
   p^.frere := nil;
   arbre_of_formule := p;
end;

procedure imprime_formule(f : formule);
var p : parbre;
begin
   p := arbre_of_formule(f);
   imprime_arbre(p^);
   detruit_arbre(p);
end; { imprime_formule }

	
                     {+------------------------+
                      |  Duplique une formule  |
                      +------------------------+}

function copie_formule(f : pformule) : pformule;
var p :  pformule;
begin
   new(p);
   p^ := f^;
   with p^ do case f^.nature of
     Constante : ;
     Variable  : ;
     Non       : p^.n := copie_formule(n);
     Bin       : begin g := copie_formule(g); d := copie_formule(d) end;
   end; { case }
   copie_formule := p;
end;